# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the test report functionality
"""

from unittest import TestCase
import os
from pathlib import Path
from xml.etree import ElementTree
from vunit.test.report import TestReport, PASSED, SKIPPED, FAILED
from vunit.ui.common import TEST_OUTPUT_PATH
from vunit.ui.results import Results


class TestTestReport(TestCase):
    """
    Test the TestReport class
    """

    def setUp(self):
        self.printer = StubPrinter()

        self.output_file_contents = 'Output file contents\n<xml>&13!--"<\\xml>'
        self.output_file_name = str(Path(__file__).parent / "test_report_output.txt")
        with open(self.output_file_name, "w") as fwrite:
            fwrite.write(self.output_file_contents)

    def report_to_str(self, report):
        """
        Helper function to create a string with color tags of the report
        """

        self.printer.reset()
        report.print_str()
        return self.printer.report_str

    def test_report_with_all_passed_tests(self):
        report = self._report_with_all_passed_tests()
        report.set_real_total_time(1.0)
        self.assertEqual(
            self.report_to_str(report),
            """\
==== Summary ========================
{gi}pass{x} passed_test0 (1.0 seconds)
{gi}pass{x} passed_test1 (2.0 seconds)
=====================================
{gi}pass{x} 2 of 2
=====================================
Total time was 3.0 seconds
Elapsed time was 1.0 seconds
=====================================
{gi}All passed!{x}
""",
        )
        self.assertTrue(report.all_ok())
        self.assertTrue(report.result_of("passed_test0").passed)
        self.assertTrue(report.result_of("passed_test1").passed)
        self.assertRaises(KeyError, report.result_of, "invalid_test")

    def test_report_with_missing_tests(self):
        report = self._report_with_missing_tests()
        report.set_real_total_time(1.0)
        self.assertEqual(
            self.report_to_str(report),
            """\
==== Summary ========================
{gi}pass{x} passed_test0 (1.0 seconds)
{gi}pass{x} passed_test1 (2.0 seconds)
=====================================
{gi}pass{x} 2 of 2
=====================================
Total time was 3.0 seconds
Elapsed time was 1.0 seconds
=====================================
{gi}All passed!{x}
{rgi}WARNING: Test execution aborted after running 2 out of 3 tests{x}
""",
        )
        self.assertTrue(report.all_ok())
        self.assertTrue(report.result_of("passed_test0").passed)
        self.assertTrue(report.result_of("passed_test1").passed)
        self.assertRaises(KeyError, report.result_of, "invalid_test")

    def test_report_with_failed_tests(self):
        report = self._report_with_some_failed_tests()
        report.set_real_total_time(12.0)
        self.assertEqual(
            self.report_to_str(report),
            """\
==== Summary ========================
{gi}pass{x} passed_test  (2.0 seconds)
{ri}fail{x} failed_test0 (11.1 seconds)
{ri}fail{x} failed_test1 (3.0 seconds)
=====================================
{gi}pass{x} 1 of 3
{ri}fail{x} 2 of 3
=====================================
Total time was 16.1 seconds
Elapsed time was 12.0 seconds
=====================================
{ri}Some failed!{x}
""",
        )
        self.assertFalse(report.all_ok())
        self.assertTrue(report.result_of("passed_test").passed)
        self.assertTrue(report.result_of("failed_test0").failed)
        self.assertTrue(report.result_of("failed_test1").failed)

    def test_report_with_skipped_tests(self):
        report = self._report_with_some_skipped_tests()
        report.set_real_total_time(3.0)
        self.assertEqual(
            self.report_to_str(report),
            """\
==== Summary ========================
{gi}pass{x} passed_test  (1.0 seconds)
{rgi}skip{x} skipped_test (0.0 seconds)
{ri}fail{x} failed_test  (3.0 seconds)
=====================================
{gi}pass{x} 1 of 3
{rgi}skip{x} 1 of 3
{ri}fail{x} 1 of 3
=====================================
Total time was 4.0 seconds
Elapsed time was 3.0 seconds
=====================================
{ri}Some failed!{x}
""",
        )
        self.assertFalse(report.all_ok())
        self.assertTrue(report.result_of("passed_test").passed)
        self.assertTrue(report.result_of("skipped_test").skipped)
        self.assertTrue(report.result_of("failed_test").failed)

    def test_report_with_no_tests(self):
        report = self._new_report()
        self.assertEqual(
            self.report_to_str(report),
            """\
{rgi}No tests were run!{x}
""",
        )
        self.assertTrue(report.all_ok())

    def assert_has_test(
        self,
        root,
        name,
        time,
        status,  # pylint: disable=too-many-arguments
        output=None,
        fmt="jenkins",
    ):  # pylint: disable=too-many-arguments
        """
        Assert that junit report xml tree contains a test
        """
        output = self.output_file_contents if output is None else output
        for test in root.findall("testcase"):
            if test.attrib["name"] == name:
                self.assertEqual(test.attrib["time"], time)
                if status == "passed":
                    self.assertEqual(len(test.findall("*")), 1)
                    self.assertEqual(len(test.findall("system-out")), 1)
                elif status == "skipped":
                    self.assertEqual(len(test.findall("*")), 2)
                    self.assertEqual(len(test.findall("skipped")), 1)
                    self.assertEqual(len(test.findall("system-out")), 1)
                elif status == "failed":
                    self.assertEqual(len(test.findall("*")), 2)
                    self.assertEqual(len(test.findall("failure")), 1)
                    self.assertEqual(len(test.findall("system-out")), 1)
                else:
                    assert False

                if status == "failed" and fmt == "bamboo":
                    self.assertEqual(test.find("failure").text, output)
                else:
                    self.assertEqual(test.find("system-out").text, output)
                break
        else:
            assert False

    def test_junit_report_with_all_passed_tests(self):
        report = self._report_with_all_passed_tests()
        root = ElementTree.fromstring(report.to_junit_xml_str())
        self.assertEqual(root.tag, "testsuite")
        self.assertEqual(len(root.findall("*")), 2)
        self.assert_has_test(root, "passed_test0", time="1.0", status="passed")
        self.assert_has_test(root, "passed_test1", time="2.0", status="passed")

    def test_junit_report_with__with_missing_output_file(self):
        report = self._report_with_all_passed_tests()
        os.remove(self.output_file_name)
        fail_output = "Failed to read output file: %s" % self.output_file_name
        root = ElementTree.fromstring(report.to_junit_xml_str())
        self.assertEqual(root.tag, "testsuite")
        self.assertEqual(len(root.findall("*")), 2)
        self.assert_has_test(root, "passed_test0", time="1.0", status="passed", output=fail_output)
        self.assert_has_test(root, "passed_test1", time="2.0", status="passed", output=fail_output)

    def test_junit_report_with_some_failed_tests(self):
        report = self._report_with_some_failed_tests()
        root = ElementTree.fromstring(report.to_junit_xml_str())
        self.assertEqual(root.tag, "testsuite")
        self.assertEqual(len(root.findall("*")), 3)
        self.assert_has_test(root, "failed_test0", time="11.1", status="failed")
        self.assert_has_test(root, "passed_test", time="2.0", status="passed")
        self.assert_has_test(root, "failed_test1", time="3.0", status="failed")

    def test_junit_report_with_some_failed_tests_bamboo_fmt(self):
        report = self._report_with_some_failed_tests()
        root = ElementTree.fromstring(report.to_junit_xml_str(xunit_xml_format="bamboo"))
        self.assertEqual(root.tag, "testsuite")
        self.assertEqual(len(root.findall("*")), 3)
        self.assert_has_test(root, "failed_test0", time="11.1", status="failed", fmt="bamboo")
        self.assert_has_test(root, "passed_test", time="2.0", status="passed", fmt="bamboo")
        self.assert_has_test(root, "failed_test1", time="3.0", status="failed", fmt="bamboo")

    def test_junit_report_with_some_skipped_tests(self):
        report = self._report_with_some_skipped_tests()
        root = ElementTree.fromstring(report.to_junit_xml_str())
        self.assertEqual(root.tag, "testsuite")
        self.assertEqual(len(root.findall("*")), 3)
        self.assert_has_test(root, "skipped_test", time="0.0", status="skipped")
        self.assert_has_test(root, "passed_test", time="1.0", status="passed")
        self.assert_has_test(root, "failed_test", time="3.0", status="failed")

    def test_junit_report_with_testcase_classname(self):
        report = self._new_report()
        report.add_result("test", PASSED, time=1.0, output_file_name=self.output_file_name)
        report.add_result("lib.entity", PASSED, time=1.0, output_file_name=self.output_file_name)
        report.add_result("lib.entity.test", PASSED, time=1.0, output_file_name=self.output_file_name)
        report.add_result(
            "lib.entity.config.test",
            PASSED,
            time=1.0,
            output_file_name=self.output_file_name,
        )
        root = ElementTree.fromstring(report.to_junit_xml_str())
        names = set(
            (elem.attrib.get("classname", None), elem.attrib.get("name", None)) for elem in root.findall("testcase")
        )
        self.assertEqual(
            names,
            set(
                [
                    (None, "test"),
                    ("lib", "entity"),
                    ("lib.entity", "test"),
                    ("lib.entity.config", "test"),
                ]
            ),
        )

    def test_dict_report_with_all_passed_tests(self):
        opath = Path(self.output_file_name).parent.parent
        test_path = opath / TEST_OUTPUT_PATH / "unit"
        output_file_name = test_path / Path(self.output_file_name).name
        results = Results(opath, None, self._report_with_all_passed_tests(output_file_name))
        report = results.get_report()
        for _, test in report.tests.items():
            self.assertEqual(test.path.name, test.relpath)
        test0 = report.tests["passed_test0"]
        test1 = report.tests["passed_test1"]
        self.assertEqual(
            {
                "passed_test0": {
                    "status": test0.status,
                    "time": test0.time,
                    "path": test0.path,
                },
                "passed_test1": {
                    "status": test1.status,
                    "time": test1.time,
                    "path": test1.path,
                },
            },
            {
                "passed_test0": {"status": "passed", "time": 1.0, "path": test_path},
                "passed_test1": {"status": "passed", "time": 2.0, "path": test_path},
            },
        )

    def _report_with_all_passed_tests(self, output_file_name=None):
        "@returns A report with all passed tests"
        if not output_file_name:
            output_file_name = self.output_file_name
        report = self._new_report()
        report.add_result("passed_test0", PASSED, time=1.0, output_file_name=output_file_name)
        report.add_result("passed_test1", PASSED, time=2.0, output_file_name=output_file_name)
        report.set_expected_num_tests(2)
        return report

    def _report_with_missing_tests(self):
        "@returns A report with all passed tests"
        report = self._new_report()
        report.add_result("passed_test0", PASSED, time=1.0, output_file_name=self.output_file_name)
        report.add_result("passed_test1", PASSED, time=2.0, output_file_name=self.output_file_name)
        report.set_expected_num_tests(3)
        return report

    def _report_with_some_failed_tests(self):
        "@returns A report with some failed tests"
        report = self._new_report()
        report.add_result("failed_test0", FAILED, time=11.12, output_file_name=self.output_file_name)
        report.add_result("passed_test", PASSED, time=2.0, output_file_name=self.output_file_name)
        report.add_result("failed_test1", FAILED, time=3.0, output_file_name=self.output_file_name)
        report.set_expected_num_tests(3)
        return report

    def _report_with_some_skipped_tests(self):
        "@returns A report with some skipped tests"
        report = self._new_report()
        report.add_result("passed_test", PASSED, time=1.0, output_file_name=self.output_file_name)
        report.add_result("skipped_test", SKIPPED, time=0.0, output_file_name=self.output_file_name)
        report.add_result("failed_test", FAILED, time=3.0, output_file_name=self.output_file_name)
        report.set_expected_num_tests(3)
        return report

    def _new_report(self):
        return TestReport(self.printer)


class StubPrinter(object):
    """
    A stub of a ColorPrinter
    """

    def __init__(self):
        self.report_str = ""

    def reset(self):
        self.report_str = ""

    def write(self, text, fg=None, bg=None):
        """
        ColorPrinter write stub
        """
        if fg is not None or bg is not None:
            self.report_str += "{"

        if fg is not None:
            self.report_str += fg

        if bg is not None:
            self.report_str += bg.upper()

        if fg is not None or bg is not None:
            self.report_str += "}"

        self.report_str += text

        if fg is not None or bg is not None:
            self.report_str += "{x}"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Check that the Python version used is supported by VUnit
"""

import sys

MAJOR = 3
MINOR = 6


def version_is_ok():
    """
    Returns true if version is high enough
    """
    version = (sys.version_info[0], sys.version_info[1])
    return version >= (MAJOR, MINOR)


if not version_is_ok():
    print(
        f"Your Python version ({sys.version_info[0]}.{sys.version_info[1]}) is too old for VUnit. "
        "Please consider upgrading."
    )
    print("VUnit supports versions:")
    print(f" - Python {MAJOR}.{MINOR} or higher")
    sys.exit(1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=too-many-public-methods, too-many-lines

"""
Test the Incisive interface
"""


import unittest
from pathlib import Path
import os
from shutil import rmtree
from unittest import mock
from vunit.sim_if.incisive import IncisiveInterface
from vunit.project import Project
from vunit.ostools import renew_path, write_file, read_file
from vunit.test.bench import Configuration
from vunit.vhdl_standard import VHDL


class TestIncisiveInterface(unittest.TestCase):
    """
    Test the Incisive interface
    """

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_vhdl_2008(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("2008"))
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_vhdl_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-v200x -extv200x",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_vhdl_file_lib.log"),
                "-quiet",
                '-nclibdirname "."',
                "-makelib lib_path",
                '"file.vhd"',
                "-endlib",
            ],
        )

        self.assertEqual(
            read_file(str(Path(self.output_path) / "cds.lib")),
            """\
## cds.lib: Defines the locations of compiled libraries.
softinclude cds_root_irun/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
# NOTE: 'virtuoso' executable not found!
# define basic ".../tools/dfII/etc/cdslib/basic"
define lib "lib_path"
define work "%s/libraries/work"
"""
            % self.output_path,
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_vhdl_2002(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("2002"))
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_vhdl_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-v200x -extv200x",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_vhdl_file_lib.log"),
                "-quiet",
                '-nclibdirname "."',
                "-makelib lib_path",
                '"file.vhd"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_vhdl_93(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl", vhdl_standard=VHDL.standard("93"))
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_vhdl_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-v93",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_vhdl_file_lib.log"),
                "-quiet",
                '-nclibdirname "."',
                "-makelib lib_path",
                '"file.vhd"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_vhdl_extra_flags(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        source_file = project.add_source_file("file.vhd", "lib", file_type="vhdl")
        source_file.set_compile_option("incisive.irun_vhdl_flags", ["custom", "flags"])
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_vhdl_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-v200x -extv200x",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_vhdl_file_lib.log"),
                "-quiet",
                "custom",
                "flags",
                '-nclibdirname "."',
                "-makelib lib_path",
                '"file.vhd"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_vhdl_hdlvar(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path, hdlvar="custom_hdlvar")
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl")
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_vhdl_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-v200x -extv200x",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-hdlvar "custom_hdlvar"',
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_vhdl_file_lib.log"),
                "-quiet",
                '-nclibdirname "."',
                "-makelib lib_path",
                '"file.vhd"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_verilog(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog")
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                '-nclibdirname "."',
                "-makelib lib",
                '"file.v"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_system_verilog(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.sv", "")
        project.add_source_file("file.sv", "lib", file_type="systemverilog")
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                '-nclibdirname "."',
                "-makelib lib",
                '"file.sv"',
                "-endlib",
            ],
        )

        self.assertEqual(
            read_file(str(Path(self.output_path) / "cds.lib")),
            """\
## cds.lib: Defines the locations of compiled libraries.
softinclude cds_root_irun/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
# NOTE: 'virtuoso' executable not found!
# define basic ".../tools/dfII/etc/cdslib/basic"
define lib "lib_path"
define work "%s/libraries/work"
"""
            % self.output_path,
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_verilog_extra_flags(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        source_file = project.add_source_file("file.v", "lib", file_type="verilog")
        source_file.set_compile_option("incisive.irun_verilog_flags", ["custom", "flags"])
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                "custom",
                "flags",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                '-nclibdirname "."',
                "-makelib lib",
                '"file.v"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_verilog_include(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", include_dirs=["include"])
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "include"',
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                '-nclibdirname "."',
                "-makelib lib",
                '"file.v"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_verilog_define(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", defines=dict(defname="defval"))
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                "-define defname=defval",
                '-nclibdirname "."',
                "-makelib lib",
                '"file.v"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.check_output", autospec=True, return_value="")
    def test_compile_project_verilog_hdlvar(self, check_output, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path, hdlvar="custom_hdlvar")
        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.v", "")
        project.add_source_file("file.v", "lib", file_type="verilog", defines=dict(defname="defval"))
        simif.compile_project(project)
        args_file = str(Path(self.output_path) / "irun_compile_verilog_file_lib.args")
        check_output.assert_called_once_with([str(Path("prefix") / "irun"), "-f", args_file], env=simif.get_env())
        self.assertEqual(
            read_file(args_file).splitlines(),
            [
                "-compile",
                "-nocopyright",
                "-licqueue",
                "-nowarn UEXPSC",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-work work",
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-hdlvar "custom_hdlvar"',
                '-log "%s"' % str(Path(self.output_path) / "irun_compile_verilog_file_lib.log"),
                "-quiet",
                '-incdir "cds_root_irun/tools/spectre/etc/ahdl/"',
                "-define defname=defval",
                '-nclibdirname "."',
                "-makelib lib",
                '"file.v"',
                "-endlib",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    def test_create_cds_lib(self, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        IncisiveInterface(prefix="prefix", output_path=self.output_path)
        self.assertEqual(
            read_file(str(Path(self.output_path) / "cds.lib")),
            """\
## cds.lib: Defines the locations of compiled libraries.
softinclude cds_root_irun/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
# NOTE: 'virtuoso' executable not found!
# define basic ".../tools/dfII/etc/cdslib/basic"
define work "%s/libraries/work"
"""
            % self.output_path,
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    def test_create_cds_lib_virtuoso(self, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = "cds_root_virtuoso"
        IncisiveInterface(prefix="prefix", output_path=self.output_path)
        self.assertEqual(
            read_file(str(Path(self.output_path) / "cds.lib")),
            """\
## cds.lib: Defines the locations of compiled libraries.
softinclude cds_root_irun/tools/inca/files/cds.lib
# needed for referencing the library 'basic' for cells 'cds_alias', 'cds_thru' etc. in analog models:
define basic "cds_root_virtuoso/tools/dfII/etc/cdslib/basic"
define work "%s/libraries/work"
"""
            % self.output_path,
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_vhdl(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)

        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl")

        with mock.patch("vunit.sim_if.check_output", autospec=True, return_value="") as dummy:
            simif.compile_project(project)

        config = make_config()
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

        self.assertEqual(
            read_file(elaborate_args_file).splitlines(),
            [
                "-elaborate",
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_elaborate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +r",
                '-input "@run"',
                "-top lib.ent:arch",
            ],
        )

        self.assertEqual(
            read_file(simulate_args_file).splitlines(),
            [
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_simulate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +r",
                '-input "@run"',
                "-top lib.ent:arch",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_verilog(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)

        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl")

        with mock.patch("vunit.sim_if.check_output", autospec=True, return_value="") as dummy:
            simif.compile_project(project)

        config = make_config(verilog=True)
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

        self.assertEqual(
            read_file(elaborate_args_file).splitlines(),
            [
                "-elaborate",
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_elaborate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +r",
                '-input "@run"',
                "-top lib.modulename:sv",
            ],
        )

        self.assertEqual(
            read_file(simulate_args_file).splitlines(),
            [
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_simulate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +r",
                '-input "@run"',
                "-top lib.modulename:sv",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_extra_flags(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        config = make_config(sim_options={"incisive.irun_sim_flags": ["custom", "flags"]})
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

        args = read_file(elaborate_args_file).splitlines()
        self.assertIn("custom", args)
        self.assertIn("flags", args)

        args = read_file(simulate_args_file).splitlines()
        self.assertIn("custom", args)
        self.assertIn("flags", args)

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_generics_and_parameters(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        config = make_config(verilog=True, generics={"genstr": "genval", "genint": 1, "genbool": True})
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

        for args_file in [elaborate_args_file, simulate_args_file]:
            args = read_file(args_file).splitlines()
            self.assertIn('-gpg "modulename.genstr => \\"genval\\""', args)
            self.assertIn('-gpg "modulename.genint => 1"', args)
            self.assertIn('-gpg "modulename.genbool => \\"True\\""', args)

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_hdlvar(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path, hdlvar="custom_hdlvar")
        config = make_config()
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

        for args_file in [elaborate_args_file, simulate_args_file]:
            args = read_file(args_file).splitlines()
            self.assertIn('-hdlvar "custom_hdlvar"', args)

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_elaborate(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        config = make_config(verilog=True)
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config, elaborate_only=True))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                )
            ]
        )

        self.assertEqual(
            read_file(elaborate_args_file).splitlines(),
            [
                "-elaborate",
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_elaborate.log"),
                "-quiet",
                "-access +r",
                '-input "@run"',
                "-top lib.modulename:sv",
            ],
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=False)
    def test_elaborate_fail(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        config = make_config()
        self.assertFalse(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                )
            ]
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, side_effect=[True, False])
    def test_simulate_fail(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None
        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path)
        config = make_config()
        self.assertFalse(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )

    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_virtuoso")
    @mock.patch("vunit.sim_if.incisive.IncisiveInterface.find_cds_root_irun")
    @mock.patch("vunit.sim_if.incisive.run_command", autospec=True, return_value=True)
    def test_simulate_gui(self, run_command, find_cds_root_irun, find_cds_root_virtuoso):
        find_cds_root_irun.return_value = "cds_root_irun"
        find_cds_root_virtuoso.return_value = None

        project = Project()
        project.add_library("lib", "lib_path")
        write_file("file.vhd", "")
        project.add_source_file("file.vhd", "lib", file_type="vhdl")

        simif = IncisiveInterface(prefix="prefix", output_path=self.output_path, gui=True)
        with mock.patch("vunit.sim_if.check_output", autospec=True, return_value="") as dummy:
            simif.compile_project(project)
        config = make_config()
        self.assertTrue(simif.simulate("suite_output_path", "test_suite_name", config))
        elaborate_args_file = str(Path("suite_output_path") / simif.name / "irun_elaborate.args")
        simulate_args_file = str(Path("suite_output_path") / simif.name / "irun_simulate.args")
        run_command.assert_has_calls(
            [
                mock.call(
                    [
                        str(Path("prefix") / "irun"),
                        "-f",
                        Path(elaborate_args_file).name,
                    ],
                    cwd=str(Path(elaborate_args_file).parent),
                    env=simif.get_env(),
                ),
                mock.call(
                    [str(Path("prefix") / "irun"), "-f", Path(simulate_args_file).name],
                    cwd=str(Path(simulate_args_file).parent),
                    env=simif.get_env(),
                ),
            ]
        )
        self.assertEqual(
            read_file(elaborate_args_file).splitlines(),
            [
                "-elaborate",
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_elaborate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +rwc",
                "-gui",
                "-top lib.ent:arch",
            ],
        )

        self.assertEqual(
            read_file(simulate_args_file).splitlines(),
            [
                "-nocopyright",
                "-licqueue",
                "-errormax 10",
                "-nowarn WRMNZD",
                "-nowarn DLCPTH",
                "-nowarn DLCVAR",
                "-ncerror EVBBOL",
                "-ncerror EVBSTR",
                "-ncerror EVBNAT",
                "-work work",
                '-nclibdirname "%s"' % str(Path(self.output_path) / "libraries"),
                '-cdslib "%s"' % str(Path(self.output_path) / "cds.lib"),
                '-log "%s"' % str(Path("suite_output_path") / simif.name / "irun_simulate.log"),
                "-quiet",
                '-reflib "lib_path"',
                "-access +rwc",
                "-gui",
                "-top lib.ent:arch",
            ],
        )

    def setUp(self):
        self.output_path = str(Path(__file__).parent / "test_incisive_out")
        renew_path(self.output_path)
        self.project = Project()
        self.cwd = os.getcwd()
        os.chdir(self.output_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if Path(self.output_path).exists():
            rmtree(self.output_path)


def make_config(sim_options=None, generics=None, verilog=False):
    """
    Utility to reduce boiler plate in tests
    """
    cfg = mock.Mock(spec=Configuration)
    cfg.library_name = "lib"

    if verilog:
        cfg.entity_name = "modulename"
        cfg.architecture_name = None
    else:
        cfg.entity_name = "ent"
        cfg.architecture_name = "arch"

    cfg.sim_options = {} if sim_options is None else sim_options
    cfg.generics = {} if generics is None else generics
    return cfg



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

# pylint: disable=too-many-lines

"""
Public VUnit User Interface (UI)
"""

import csv
import sys
import traceback
import logging
import json
import os
from typing import Optional, Set, Union
from pathlib import Path
from fnmatch import fnmatch
from ..database import PickledDataBase, DataBase
from .. import ostools
from ..vunit_cli import VUnitCLI
from ..sim_if.factory import SIMULATOR_FACTORY
from ..sim_if import SimulatorInterface
from ..color_printer import COLOR_PRINTER, NO_COLOR_PRINTER

from ..project import Project
from ..exceptions import CompileError
from ..location_preprocessor import LocationPreprocessor
from ..check_preprocessor import CheckPreprocessor
from ..parsing.encodings import HDL_FILE_ENCODING
from ..builtins import Builtins
from ..vhdl_standard import VHDL, VHDLStandard
from ..test.bench_list import TestBenchList
from ..test.report import TestReport
from ..test.runner import TestRunner

from .common import LOGGER, TEST_OUTPUT_PATH, select_vhdl_standard, check_not_empty
from .source import SourceFile, SourceFileList
from .library import Library
from .results import Results


class VUnit(object):  # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """
    The public interface of VUnit

    :example:

    .. code-block:: python

       from vunit import VUnit
    """

    @classmethod
    def from_argv(
        cls,
        argv=None,
        compile_builtins: Optional[bool] = True,
        vhdl_standard: Optional[str] = None,
    ):
        """
        Create VUnit instance from command line arguments.

        :param argv: Use explicit argv instead of actual command line argument
        :param compile_builtins: Do not compile builtins. Used for VUnit internal testing.
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None the VUNIT_VHDL_STANDARD environment variable is used
        :returns: A :class:`.VUnit` object instance

        :example:

        .. code-block:: python

           from vunit import VUnit
           prj = VUnit.from_argv()

        """
        args = VUnitCLI().parse_args(argv=argv)
        return cls.from_args(args, compile_builtins=compile_builtins, vhdl_standard=vhdl_standard)

    @classmethod
    def from_args(
        cls,
        args,
        compile_builtins: Optional[bool] = True,
        vhdl_standard: Optional[str] = None,
    ):
        """
        Create VUnit instance from args namespace.
        Intended for users who adds custom command line options.
        See :class:`vunit.vunit_cli.VUnitCLI` class to learn about
        adding custom command line options.

        :param args: The parsed argument namespace object
        :param compile_builtins: Do not compile builtins. Used for VUnit internal testing.
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None the VUNIT_VHDL_STANDARD environment variable is used
        :returns: A :class:`.VUnit` object instance

        """
        return cls(args, compile_builtins=compile_builtins, vhdl_standard=vhdl_standard)

    def __init__(
        self,
        args,
        compile_builtins: Optional[bool] = True,
        vhdl_standard: Optional[str] = None,
    ):
        self._args = args
        self._configure_logging(args.log_level)
        self._output_path = str(Path(args.output_path).resolve())

        if args.no_color:
            self._printer = NO_COLOR_PRINTER
        else:
            self._printer = COLOR_PRINTER

        def test_filter(name, attribute_names):
            keep = any(fnmatch(name, pattern) for pattern in args.test_patterns)

            if args.with_attributes is not None:
                keep = keep and set(args.with_attributes).issubset(attribute_names)

            if args.without_attributes is not None:
                keep = keep and set(args.without_attributes).isdisjoint(attribute_names)
            return keep

        self._test_filter = test_filter
        self._vhdl_standard: VHDLStandard = select_vhdl_standard(vhdl_standard)

        self._external_preprocessors = []  # type: ignore
        self._location_preprocessor = None
        self._check_preprocessor = None

        self._simulator_class = SIMULATOR_FACTORY.select_simulator()

        # Use default simulator options if no simulator was present
        if self._simulator_class is None:
            simulator_class = SimulatorInterface
            self._simulator_output_path = str(Path(self._output_path) / "none")
        else:
            simulator_class = self._simulator_class
            self._simulator_output_path = str(Path(self._output_path) / simulator_class.name)

        self._create_output_path(args.clean)

        database = self._create_database()
        self._project = Project(
            database=database,
            depend_on_package_body=simulator_class.package_users_depend_on_bodies,
        )

        self._test_bench_list = TestBenchList(database=database)

        self._builtins = Builtins(self, self._vhdl_standard, simulator_class)
        if compile_builtins:
            self.add_builtins()

    def _create_database(self):
        """
        Create a persistent database to store expensive parse results

        Check for Python version used to create the database is the
        same as the running python instance or re-create
        """
        project_database_file_name = str(Path(self._output_path) / "project_database")
        create_new = False
        key = b"version"
        version = str((9, sys.version)).encode()
        database = None
        try:
            database = DataBase(project_database_file_name)
            create_new = (key not in database) or (database[key] != version)
        except KeyboardInterrupt as exk:
            raise KeyboardInterrupt from exk
        except:  # pylint: disable=bare-except
            traceback.print_exc()
            create_new = True

        if create_new:
            database = DataBase(project_database_file_name, new=True)
        database[key] = version

        return PickledDataBase(database)

    @staticmethod
    def _configure_logging(log_level):
        """
        Configure logging based on log_level string
        """
        level = getattr(logging, log_level.upper())
        logging.basicConfig(filename=None, format="%(levelname)7s - %(message)s", level=level)

    def _which_vhdl_standard(self, vhdl_standard: Optional[str]) -> VHDLStandard:
        """
        Return default vhdl_standard if the argument is None
        The argument is a string from the user
        """
        if vhdl_standard is None:
            return self._vhdl_standard

        return VHDL.standard(vhdl_standard)

    def add_external_library(self, library_name, path: Union[str, Path], vhdl_standard: Optional[str] = None):
        """
        Add an externally compiled library as a black-box

        :param library_name: The name of the external library
        :param path: The path to the external library directory
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None the VUNIT_VHDL_STANDARD environment variable is used
        :returns: The created :class:`.Library` object

        :example:

        .. code-block:: python

           prj.add_external_library("unisim", "path/to/unisim/")

        """

        self._project.add_library(
            library_name,
            Path(path).resolve(),
            self._which_vhdl_standard(vhdl_standard),
            is_external=True,
        )
        return self.library(library_name)

    def add_source_files_from_csv(self, project_csv_path: Union[str, Path], vhdl_standard: Optional[str] = None):
        """
        Add a project configuration, mapping all the libraries and files

        :param project_csv_path: path to csv project specification, each line contains the name
                                 of the library and the path to one file 'lib_name,filename'
                                 note that all filenames are relative to the parent folder of the
                                 csv file
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None, the VUNIT_VHDL_STANDARD environment variable is used
        :returns: A list of files (:class `.SourceFileList`) that were added

        """
        libs: Set[str] = set()
        files = SourceFileList([])

        ppath = Path(project_csv_path)

        with ppath.open("r", encoding="utf-8") as csv_path_file:
            for row in csv.reader(csv_path_file):
                if len(row) == 2:
                    lib_name = row[0].strip()
                    no_normalized_file = row[1].strip()
                    file_name_ = str((ppath.parent / no_normalized_file).resolve())
                    lib = self.library(lib_name) if lib_name in libs else self.add_library(lib_name)
                    libs.add(lib_name)
                    file_ = lib.add_source_file(file_name_, vhdl_standard=vhdl_standard)
                    files.append(file_)
                elif len(row) > 2:
                    LOGGER.error("More than one library and one file in csv description")
        return files

    def add_library(
        self,
        library_name: str,
        vhdl_standard: Optional[str] = None,
        allow_duplicate: Optional[bool] = False,
    ):
        """
        Add a library managed by VUnit.

        :param library_name: The name of the library
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None the VUNIT_VHDL_STANDARD environment variable is used
        :param allow_duplicate: Set to True to allow the same library
                                to be added multiple times. Subsequent additions will just
                                return the previously created library.
        :returns: The created :class:`.Library` object

        :example:

        .. code-block:: python

           library = prj.add_library("lib")

        """
        standard = self._which_vhdl_standard(vhdl_standard)
        path = Path(self._simulator_output_path) / "libraries" / library_name
        if not self._project.has_library(library_name):
            self._project.add_library(library_name, str(path.resolve()), standard)
        elif not allow_duplicate:
            raise ValueError(f"Library {library_name!s} already added. Use allow_duplicate to ignore this error.")
        return self.library(library_name)

    def library(self, library_name: str):
        """
        Get a library

        :param library_name: The name of the library
        :returns: A :class:`.Library` object
        """
        if not self._project.has_library(library_name):
            raise KeyError(library_name)
        return Library(library_name, self, self._project, self._test_bench_list)

    def set_attribute(self, name: str, value: str, allow_empty: Optional[bool] = False):
        """
        Set a value of attribute in all |configurations|

        :param name: The name of the attribute
        :param value: The value of the attribute
        :param allow_empty: To disable an error when no test benches were found

        :example:

        .. code-block:: python

           prj.set_attribute(".foo", "bar")

        .. note::
           Only affects test benches added *before* the attribute is set.
        """
        test_benches = self._test_bench_list.get_test_benches()
        for test_bench in check_not_empty(test_benches, allow_empty, "No test benches found"):
            test_bench.set_attribute(name, value)

    def set_generic(self, name: str, value: str, allow_empty: Optional[bool] = False):
        """
        Set a value of generic in all |configurations|

        :param name: The name of the generic
        :param value: The value of the generic
        :param allow_empty: To disable an error when no test benches were found

        :example:

        .. code-block:: python

           prj.set_generic("data_width", 16)

        .. note::
           Only affects test benches added *before* the generic is set.
        """
        test_benches = self._test_bench_list.get_test_benches()
        for test_bench in check_not_empty(test_benches, allow_empty, "No test benches found"):
            test_bench.set_generic(name.lower(), value)

    def set_parameter(self, name: str, value: str, allow_empty: Optional[bool] = False):
        """
        Set value of parameter in all |configurations|

        :param name: The name of the parameter
        :param value: The value of the parameter
        :param allow_empty: To disable an error when no test benches were found

        :example:

        .. code-block:: python

           prj.set_parameter("data_width", 16)

        .. note::
           Only affects test benches added *before* the parameter is set.
        """
        test_benches = self._test_bench_list.get_test_benches()
        for test_bench in check_not_empty(test_benches, allow_empty, "No test benches found"):
            test_bench.set_generic(name, value)

    def set_sim_option(
        self,
        name: str,
        value: str,
        allow_empty: Optional[bool] = False,
        overwrite: Optional[bool] = True,
    ):
        """
        Set simulation option in all |configurations|

        :param name: |simulation_options|
        :param value: The value of the simulation option
        :param allow_empty: To disable an error when no test benches were found
        :param overwrite: To overwrite the option or append to the existing value

        :example:

        .. code-block:: python

           prj.set_sim_option("ghdl.a_flags", ["--no-vital-checks"])

        .. note::
           Only affects test benches added *before* the option is set.
        """
        test_benches = self._test_bench_list.get_test_benches()
        for test_bench in check_not_empty(test_benches, allow_empty, "No test benches found"):
            test_bench.set_sim_option(name, value, overwrite)

    def set_compile_option(self, name: str, value: str, allow_empty: Optional[bool] = False):
        """
        Set compile option of all files

        :param name: |compile_option|
        :param value: The value of the compile option
        :param allow_empty: To disable an error when no source files were found

        :example:

        .. code-block:: python

           prj.set_compile_option("ghdl.a_flags", ["--no-vital-checks"])


        .. note::
           Only affects files added *before* the option is set.
        """
        source_files = self._project.get_source_files_in_order()
        for source_file in check_not_empty(source_files, allow_empty, "No source files found"):
            source_file.set_compile_option(name, value)

    def add_compile_option(self, name: str, value: str, allow_empty: Optional[bool] = False):
        """
        Add compile option to all files

        :param name: |compile_option|
        :param value: The value of the compile option
        :param allow_empty: To disable an error when no source files were found

        .. note::
           Only affects files added *before* the option is set.
        """
        source_files = self._project.get_source_files_in_order()
        for source_file in check_not_empty(source_files, allow_empty, "No source files found"):
            source_file.add_compile_option(name, value)

    def get_source_file(self, file_name: Union[str, Path], library_name: Optional[str] = None):
        """
        Get a source file

        :param file_name: The name of the file as a relative or absolute path
        :param library_name: The name of a specific library to search if not all libraries
        :returns: A :class:`.SourceFile` object
        """

        fstr = str(file_name)

        files = self.get_source_files(fstr, library_name, allow_empty=True)
        if len(files) > 1:
            raise ValueError(f"Found file named '{fstr!s}' in multiple-libraries, " "add explicit library_name.")
        if not files:
            if library_name is None:
                raise ValueError(f"Found no file named '{fstr!s}'")

            raise ValueError(f"Found no file named '{fstr!s}' in library '{library_name!s}'")
        return files[0]

    def get_source_files(
        self,
        pattern="*",
        library_name: Optional[str] = None,
        allow_empty: Optional[bool] = False,
    ):
        """
        Get a list of source files

        :param pattern: A wildcard pattern matching either an absolute or relative path
        :param library_name: The name of a specific library to search if not all libraries
        :param allow_empty: To disable an error if no files matched the pattern
        :returns: A :class:`.SourceFileList` object
        """
        results = []
        for source_file in self._project.get_source_files_in_order():
            if library_name is not None:
                if source_file.library.name != library_name:
                    continue

            if not (
                fnmatch(str(Path(source_file.name).resolve()), pattern)
                or fnmatch(ostools.simplify_path(source_file.name), pattern)
            ):
                continue

            results.append(SourceFile(source_file, self._project, self))

        check_not_empty(
            results,
            allow_empty,
            f"Pattern {pattern!r} did not match any file"
            + (f"within library {library_name!s}" if library_name is not None else ""),
        )

        return SourceFileList(results)

    def add_source_files(  # pylint: disable=too-many-arguments
        self,
        pattern,
        library_name: str,
        preprocessors=None,
        include_dirs=None,
        defines=None,
        allow_empty: Optional[bool] = False,
        vhdl_standard: Optional[str] = None,
        no_parse: Optional[bool] = False,
        file_type=None,
    ):
        """
        Add source files matching wildcard pattern to library

        :param pattern: A wildcard pattern matching the files to add or a list of files
        :param library_name: The name of the library to add files into
        :param include_dirs: A list of include directories
        :param defines: A dictionary containing Verilog defines to be set
        :param allow_empty: To disable an error if no files matched the pattern
        :param vhdl_standard: The VHDL standard used to compile files,
                              if None VUNIT_VHDL_STANDARD environment variable is used
        :param no_parse: Do not parse file(s) for dependency or test scanning purposes
        :param file_type: The type of the file; ``"vhdl"``, ``"verilog"``  or ``"systemverilog"``.
                          Auto-detected by default when set to ``None``.
        :returns: A list of files (:class:`.SourceFileList`) which were added

        :example:

        .. code-block:: python

           prj.add_source_files("*.vhd", "lib")

        """
        return self.library(library_name).add_source_files(
            pattern=pattern,
            preprocessors=preprocessors,
            include_dirs=include_dirs,
            defines=defines,
            allow_empty=allow_empty,
            vhdl_standard=vhdl_standard,
            no_parse=no_parse,
            file_type=file_type,
        )

    def add_source_file(  # pylint: disable=too-many-arguments
        self,
        file_name: Union[str, Path],
        library_name: str,
        preprocessors=None,
        include_dirs=None,
        defines=None,
        vhdl_standard: Optional[str] = None,
        no_parse: Optional[bool] = False,
        file_type=None,
    ):
        """
        Add source file to library

        :param file_name: The name of the file
        :param library_name: The name of the library to add the file into
        :param include_dirs: A list of include directories
        :param defines: A dictionary containing Verilog defines to be set
        :param vhdl_standard: The VHDL standard used to compile this file,
                              if None VUNIT_VHDL_STANDARD environment variable is used
        :param no_parse: Do not parse file for dependency or test scanning purposes
        :param file_type: The type of the file; ``"vhdl"``, ``"verilog"``  or ``"systemverilog"``.
                          Auto-detected by default when set to ``None``.
        :returns: The :class:`.SourceFile` which was added

        :example:

        .. code-block:: python

           prj.add_source_file("file.vhd", "lib")

        """
        return self.library(library_name).add_source_file(
            file_name=str(file_name),
            preprocessors=preprocessors,
            include_dirs=include_dirs,
            defines=defines,
            vhdl_standard=vhdl_standard,
            no_parse=no_parse,
            file_type=file_type,
        )

    def _preprocess(self, library_name: str, file_name: Union[str, Path], preprocessors):
        """
        Preprocess file_name within library_name using explicit preprocessors
        if preprocessors is None then use implicit globally defined processors
        """
        # @TODO dependency checking etc...

        if preprocessors is None:
            preprocessors = [self._location_preprocessor, self._check_preprocessor]
            preprocessors = [p for p in preprocessors if p is not None]
            preprocessors = self._external_preprocessors + preprocessors

        fstr = str(file_name)

        if not preprocessors:
            return fstr

        fname = str(Path(file_name).name)

        try:
            code = ostools.read_file(file_name, encoding=HDL_FILE_ENCODING)
            for preprocessor in preprocessors:
                code = preprocessor.run(code, fname)
        except KeyboardInterrupt as exk:
            raise KeyboardInterrupt from exk
        except:  # pylint: disable=bare-except
            traceback.print_exc()
            LOGGER.error("Failed to preprocess %s", fstr)
            return fstr
        else:
            pp_file_name = str(Path(self._preprocessed_path) / library_name / fname)

            idx = 1
            while ostools.file_exists(pp_file_name):
                LOGGER.debug("Preprocessed file exists '%s', adding prefix", pp_file_name)
                pp_file_name = str(
                    Path(self._preprocessed_path) / library_name / f"{idx}_{fname!s}",
                )
                idx += 1

            ostools.write_file(pp_file_name, code, encoding=HDL_FILE_ENCODING)
            return pp_file_name

    def add_preprocessor(self, preprocessor):
        """
        Add a custom preprocessor to be used on all files, must be called before adding any files
        """
        self._external_preprocessors.append(preprocessor)

    def enable_location_preprocessing(self, additional_subprograms=None, exclude_subprograms=None):
        """
        Inserts file name and line number information into VUnit check and log subprograms calls. Custom
        subprograms can also be added. Must be called before adding any files.

        :param additional_subprograms: List of custom subprograms to add the line_num and file_name parameters to.
        :param exclude_subprograms: List of VUnit subprograms to exclude from location preprocessing. Used to \
avoid location preprocessing of other functions sharing name with a VUnit log or check subprogram.

        :example:

        .. code-block:: python

           prj.enable_location_preprocessing(additional_subprograms=['my_check'],
                                             exclude_subprograms=['log'])

        """
        preprocessor = LocationPreprocessor()
        if additional_subprograms is not None:
            for subprogram in additional_subprograms:
                preprocessor.add_subprogram(subprogram)

        if exclude_subprograms is not None:
            for subprogram in exclude_subprograms:
                preprocessor.remove_subprogram(subprogram)
        self._location_preprocessor = preprocessor

    def enable_check_preprocessing(self):
        """
        Inserts error context information into VUnit check_relation calls
        """
        self._check_preprocessor = CheckPreprocessor()

    def main(self, post_run=None):
        """
        Run vunit main function and exit

        :param post_run: A callback function which is called after
          running tests. The function must accept a single `results`
          argument which is an instance of :class:`.Results`
        """
        try:
            all_ok = self._main(post_run)
        except KeyboardInterrupt:
            sys.exit(1)
        except CompileError:
            sys.exit(1)
        except SystemExit:
            sys.exit(1)
        except:  # pylint: disable=bare-except
            if self._args.dont_catch_exceptions:
                raise
            traceback.print_exc()
            sys.exit(1)

        if (not all_ok) and (not self._args.exit_0):
            sys.exit(1)

        sys.exit(0)

    def _create_tests(self, simulator_if: Union[None, SimulatorInterface]):
        """
        Create the test cases
        """
        self._test_bench_list.warn_when_empty()
        test_list = self._test_bench_list.create_tests(simulator_if, self._args.elaborate)
        test_list.keep_matches(self._test_filter)
        return test_list

    def _main(self, post_run):
        """
        Base vunit main function without performing exit
        """

        if self._args.export_json is not None:
            return self._main_export_json(self._args.export_json)

        if self._args.list:
            return self._main_list_only()

        if self._args.files:
            return self._main_list_files_only()

        if self._args.compile:
            return self._main_compile_only()

        all_ok = self._main_run(post_run)
        return all_ok

    def _create_simulator_if(self):
        """
        Create new simulator instance
        """

        if self._simulator_class is None:
            LOGGER.error(
                "No available simulator detected.\n"
                "Simulator binary folder must be available in PATH environment variable.\n"
                "Simulator binary folder can also be set the in VUNIT_<SIMULATOR_NAME>_PATH environment variable.\n"
            )
            sys.exit(1)

        if not Path(self._simulator_output_path).exists():
            os.makedirs(self._simulator_output_path)

        return self._simulator_class.from_args(args=self._args, output_path=self._simulator_output_path)

    def _main_run(self, post_run):
        """
        Main with running tests
        """
        simulator_if = self._create_simulator_if()
        test_list = self._create_tests(simulator_if)
        self._compile(simulator_if)
        print()

        start_time = ostools.get_time()
        report = TestReport(printer=self._printer)

        try:
            self._run_test(test_list, report)
        except KeyboardInterrupt:
            print()
            LOGGER.debug("_main: Caught Ctrl-C shutting down")
        finally:
            del test_list

        report.set_real_total_time(ostools.get_time() - start_time)
        report.print_str()

        if post_run is not None:
            post_run(results=Results(self._output_path, simulator_if, report))

        del simulator_if

        if self._args.xunit_xml is not None:
            xml = report.to_junit_xml_str(self._args.xunit_xml_format)
            ostools.write_file(self._args.xunit_xml, xml)

        return report.all_ok()

    def _main_list_only(self):
        """
        Main function when only listing test cases
        """
        test_list = self._create_tests(simulator_if=None)
        for test_name in test_list.test_names:
            print(test_name)
        print(f"Listed {test_list.num_tests} tests")
        return True

    def _main_export_json(self, json_file_name: Union[str, Path]):  # pylint: disable=too-many-locals
        """
        Main function when exporting to JSON
        """

        file_objects = self.get_compile_order()
        files = []
        for source_file in file_objects:
            files.append(
                dict(
                    file_name=str(Path(source_file.name).resolve()),
                    library_name=source_file.library.name,
                )
            )

        tests = []
        for test_suite in self._create_tests(simulator_if=None):
            test_information = test_suite.test_information
            test_configuration = test_suite.test_configuration
            for name in test_suite.test_names:
                info = test_information[name]
                config = test_configuration[name]

                attributes = {}
                for attr in info.attributes:
                    attributes[attr.name] = attr.value

                attributes.update(config.attributes)

                tests.append(
                    dict(
                        name=name,
                        location=dict(
                            file_name=str(info.location.file_name),
                            offset=info.location.offset,
                            length=info.location.length,
                        ),
                        attributes=attributes,
                    )
                )

        json_data = dict(
            # The semantic version (https://semver.org/) of the JSON export data format
            export_format_version=dict(major=1, minor=0, patch=0),
            # The set of files added to the project
            files=files,
            # The list of all tests
            tests=tests,
        )

        with Path(json_file_name).open("w", encoding="utf-8") as fptr:
            json.dump(json_data, fptr, sort_keys=True, indent=4, separators=(",", ": "))

        return True

    def _main_list_files_only(self):
        """
        Main function when only listing files
        """
        files = self.get_compile_order()
        for source_file in files:
            print(f"{source_file.library.name!s}, {source_file.name!s}")
        print(f"Listed {len(files)} files")
        return True

    def _main_compile_only(self):
        """
        Main function when only compiling
        """
        simulator_if = self._create_simulator_if()
        self._compile(simulator_if)
        return True

    def _create_output_path(self, clean: bool):
        """
        Create or re-create the output path if necessary
        """
        if clean:
            ostools.renew_path(self._output_path)
        elif not Path(self._output_path).exists():
            os.makedirs(self._output_path)

        ostools.renew_path(self._preprocessed_path)

    @property
    def vhdl_standard(self) -> str:
        return str(self._vhdl_standard)

    @property
    def _preprocessed_path(self):
        return str(Path(self._output_path) / "preprocessed")

    @property
    def codecs_path(self):
        return str(Path(self._output_path) / "codecs")

    def _compile(self, simulator_if: SimulatorInterface):
        """
        Compile entire project
        """
        # get test benches
        if self._args.minimal:
            target_files = self._get_testbench_files(simulator_if)
        else:
            target_files = None

        simulator_if.compile_project(
            self._project,
            continue_on_error=self._args.keep_compiling,
            printer=self._printer,
            target_files=target_files,
        )

    def _get_testbench_files(self, simulator_if: Union[None, SimulatorInterface]):
        """
        Return the list of all test bench files for the currently selected tests to run
        """
        test_list = self._create_tests(simulator_if)
        tb_file_names = {test_suite.file_name for test_suite in test_list}
        return [
            self.get_source_file(file_name)._source_file  # pylint: disable=protected-access
            for file_name in tb_file_names
        ]

    def _run_test(self, test_cases, report):
        """
        Run the test suites and return the report
        """

        if self._args.verbose:
            verbosity = TestRunner.VERBOSITY_VERBOSE
        elif self._args.quiet:
            verbosity = TestRunner.VERBOSITY_QUIET
        else:
            verbosity = TestRunner.VERBOSITY_NORMAL

        runner = TestRunner(
            report,
            str(Path(self._output_path) / TEST_OUTPUT_PATH),
            verbosity=verbosity,
            num_threads=self._args.num_threads,
            fail_fast=self._args.fail_fast,
            dont_catch_exceptions=self._args.dont_catch_exceptions,
            no_color=self._args.no_color,
        )
        runner.run(test_cases)

    def add_builtins(self, external=None):
        """
        Add vunit VHDL builtin libraries

        :param external: struct to provide bridges for the external VHDL API.
                         {
                             'string': ['path/to/custom/file'],
                             'integer': ['path/to/custom/file']
                         }.
        """
        self._builtins.add_vhdl_builtins(external=external)

    def add_com(self):
        """
        Add communication package
        """
        self._builtins.add("com")

    def add_array_util(self):
        """
        Add array util
        """
        self._builtins.add("array_util")

    def add_random(self):
        """
        Add random
        """
        self._builtins.add("random")

    def add_verification_components(self):
        """
        Add verification component library
        """
        self._builtins.add("verification_components")

    def add_osvvm(self):
        """
        Add osvvm library
        """
        self._builtins.add("osvvm")

    def add_json4vhdl(self):
        """
        Add JSON-for-VHDL library
        """
        self._builtins.add("json4vhdl")

    def get_compile_order(self, source_files=None):
        """
        Get the compile order of all or specific source files and
        their dependencies.

        A dependency of file **A** in terms of compile order is any
        file **B** which **must** be successfully compiled before **A**
        can be successfully compiled.

        This is **not** the same as all files required to successfully elaborate **A**.
        For example using component instantiation in VHDL there is no
        compile order dependency but the component instance will not
        elaborate if there is no binding component.

        :param source_files: A list of :class:`.SourceFile` objects or `None` meaing all
        :returns: A list of :class:`.SourceFile` objects in compile order.
        """
        if source_files is None:
            source_files = self.get_source_files(allow_empty=True)

        target_files = [source_file._source_file for source_file in source_files]  # pylint: disable=protected-access
        source_files = self._project.get_dependencies_in_compile_order(target_files)
        return SourceFileList([SourceFile(source_file, self._project, self) for source_file in source_files])

    def get_implementation_subset(self, source_files):
        """
        Get the subset of files which are required to successfully
        elaborate the list of input files without any missing
        dependencies.

        :param source_files: A list of :class:`.SourceFile` objects
        :returns: A list of :class:`.SourceFile` objects which is the implementation subset.
        """
        target_files = [source_file._source_file for source_file in source_files]  # pylint: disable=protected-access
        source_files = self._project.get_dependencies_in_compile_order(target_files, implementation_dependencies=True)
        return SourceFileList([SourceFile(source_file, self._project, self) for source_file in source_files])

    def get_simulator_name(self):
        """
        Get the name of the simulator used.

        Will return None if no simulator was found.
        """
        if self._simulator_class is None:
            return None
        return self._simulator_class.name

    def simulator_supports_coverage(self):
        """
        Returns True when the simulator supports coverage

        Will return None if no simulator was found.
        """
        if self._simulator_class is None:
            return None
        return self._simulator_class.supports_coverage()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
UI class Results
"""

from pathlib import Path
from typing import Dict, Union
from .common import TEST_OUTPUT_PATH


class Results(object):
    """
    Gives access to results after running tests
    """

    def __init__(self, output_path, simulator_if, report):
        self._output_path = output_path
        self._simulator_if = simulator_if
        self._report = report

    def merge_coverage(self, file_name, args=None):
        """
        Create a merged coverage report from the individual coverage files

        :param file_name: The resulting coverage file name.
        :param args: The tool arguments for the merge command. Should be a list of strings.
        """
        self._simulator_if.merge_coverage(file_name=file_name, args=args)

    def get_report(self):
        """
        Get a report (dictionary) of tests: status, time and output path

        :returns: A :class:`Report` object
        """
        report = Report(self._output_path)
        for test in self._report._test_results_in_order():  # pylint: disable=protected-access
            obj = test.to_dict()
            report.tests.update(
                {
                    test.name: TestResult(
                        Path(self._output_path) / TEST_OUTPUT_PATH,
                        obj["status"],
                        obj["time"],
                        obj["path"],
                    )
                }
            )
        return report


class Report(object):
    """
    Gives access to test results and paths after running tests

    :data output_path: Absolute path to the output path (see :ref:`cli`)
    :data tests: Dictionary of :class:`TestResult` objects
    """

    def __init__(self, output_path: Union[str, Path]):
        self.output_path = Path(output_path)
        self.tests: Dict[str, TestResult] = {}


class TestResult(object):
    """
    Gives access to a subset of the results of a test

    :data status: Result status (passed, failed or skipped)
    :data time: Simulation time
    :data path: Absolute path of the test output

    :example:

    .. code-block:: python

       def post_func(results):
           report = results.get_report()
           print(report.output_path)
           for key, test in report.tests.items():
               print(key)
               print(test.status)
               print(test.time)
               print(test.path)
               print(test.relpath)

       vu.main(post_run=post_func)
    """

    def __init__(self, test_output_path: Union[str, Path], status, time, path: Union[str, Path]):
        self._test_output_path = Path(test_output_path)
        self.status = status
        self.time = time
        self.path = Path(path)

    @property
    def relpath(self) -> str:
        """
        If the path is a subdir to the default TEST_OUTPUT_PATH, return the subdir only
        """
        base = self.path.name
        return str(base if (self._test_output_path / base).resolve() == self.path.resolve() else self.path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
UI class PackageFacade
"""

from pathlib import Path
from ..com import codec_generator


class PackageFacade(object):
    """
    User interface of a Package
    """

    def __init__(self, parent, library_name, package_name, design_unit):
        self._parent = parent
        self._library_name = library_name
        self._package_name = package_name
        self._design_unit = design_unit

    def generate_codecs(self, codec_package_name=None, used_packages=None, output_file_name=None):
        """
        Generates codecs for the datatypes in this Package
        """
        if codec_package_name is None:
            codec_package_name = self._package_name + "_codecs"

        if output_file_name is None:
            codecs_path = Path(self._parent.codecs_path) / self._library_name
            file_extension = Path(self._design_unit.source_file.name).suffix
            output_file_name = codecs_path / (codec_package_name + file_extension)

        codec_generator.generate_codecs(self._design_unit, codec_package_name, used_packages, output_file_name)

        return self._parent.add_source_files(output_file_name, self._library_name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

from pathlib import Path
from vunit import VUnit

ROOT = Path(__file__).parent

UI = VUnit.from_argv()
UI.add_random()
UI.library("vunit_lib").add_source_files(ROOT / "test" / "*.vhd")

UI.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Provides operating systems dependent functionality that can be easily
stubbed for testing
"""


import time
import subprocess
import threading
import shutil
from queue import Queue, Empty
from pathlib import Path
from os.path import getmtime, relpath, splitdrive
import os
from os import getcwd, makedirs
import io

import logging

LOGGER = logging.getLogger(__name__)

IS_WINDOWS_SYSTEM = os.name == "nt"


class ProgramStatus(object):
    """
    Maintain global program status to support graceful shutdown
    """

    def __init__(self):
        self._lock = threading.Lock()
        self._shutting_down = False

    @property
    def is_shutting_down(self):
        with self._lock:  # pylint: disable=not-context-manager
            return self._shutting_down

    def check_for_shutdown(self):
        if self.is_shutting_down:
            raise KeyboardInterrupt

    def shutdown(self):
        with self._lock:  # pylint: disable=not-context-manager
            LOGGER.debug("ProgramStatus.shutdown")
            self._shutting_down = True

    def reset(self):
        with self._lock:  # pylint: disable=not-context-manager
            self._shutting_down = False


PROGRAM_STATUS = ProgramStatus()


class InterruptableQueue(object):
    """
    A Queue which can be interrupted
    """

    def __init__(self):
        self._queue = Queue()

    def get(self):
        """
        Get a value from the queue
        """
        while True:
            PROGRAM_STATUS.check_for_shutdown()
            try:
                return self._queue.get(timeout=0.1)
            except Empty:
                pass

    def put(self, value):
        self._queue.put(value)

    def empty(self):
        return self._queue.empty()


class Process(object):
    """
    A simple process interface which supports asynchronously consuming the stdout and stderr
    of the process while it is running.
    """

    class NonZeroExitCode(Exception):
        pass

    def __init__(self, args, cwd=None, env=None):
        self._args = args

        # Create process with new process group
        # Sending a signal to a process group will send it to all children
        # Hopefully this way no orphaned processes will be left behind
        if IS_WINDOWS_SYSTEM:  # Windows
            self._process = subprocess.Popen(  # pylint: disable=consider-using-with
                args,
                bufsize=0,
                cwd=cwd,
                env=env,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                # Create new process group on Windows
                creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
            )
        else:
            self._process = subprocess.Popen(  # pylint: disable=consider-using-with
                args,
                bufsize=0,
                cwd=cwd,
                env=env,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                # Create new process group on POSIX, setpgrp does not exist on Windows
                preexec_fn=os.setpgrp,  # pylint: disable=no-member
            )

        LOGGER.debug("Started process with pid=%i: '%s'", self._process.pid, (" ".join(args)))

        self._queue = InterruptableQueue()
        self._reader = AsynchronousFileReader(self._process.stdout, self._queue)
        self._reader.start()

    def write(self, *args, **kwargs):
        """Write to stdin"""
        if not self._process.stdin.closed:
            self._process.stdin.write(*args, **kwargs)

    def writeline(self, line):
        """Write a line to stdin"""
        if not self._process.stdin.closed:
            self._process.stdin.write(line + "\n")
            self._process.stdin.flush()

    def next_line(self):
        """
        Return either the next line or the exit code
        """

        if not self._reader.eof():
            # Show what we received from standard output.
            msg = self._queue.get()

            if msg is not None:
                return msg

        retcode = self.wait()
        return retcode

    def wait(self):
        """
        Wait while without completely blocking to avoid
        deadlock when shutting down
        """
        while self._process.poll() is None:
            PROGRAM_STATUS.check_for_shutdown()
            time.sleep(0.05)
            LOGGER.debug("Waiting for process with pid=%i to stop", self._process.pid)
        return self._process.returncode

    def is_alive(self):
        """
        Returns true if alive
        """
        return self._process.poll() is None

    def consume_output(self, callback=print):
        """
        Consume the output of the process.
        The output is interpreted as UTF-8 text.

        @param callback Called for each line of output
        @raises Process.NonZeroExitCode when the process does not exit with code zero
        """

        def default_callback(*args, **kwargs):
            pass

        if not callback:
            callback = default_callback

        while not self._reader.eof():
            line = self._queue.get()
            if line is None:
                break

            if callback(line) is not None:
                return

        retcode = None
        while retcode is None:
            retcode = self.wait()
            if retcode != 0:
                raise Process.NonZeroExitCode

    def terminate(self):
        """
        Terminate the process
        """
        # Let's be tidy and join the threads we've started.
        if self._process.poll() is None:
            LOGGER.debug("Terminating process with pid=%i", self._process.pid)
            self._process.terminate()

        if self._process.poll() is None:
            time.sleep(0.05)

        if self._process.poll() is None:
            LOGGER.debug("Killing process with pid=%i", self._process.pid)
            self._process.kill()

        if self._process.poll() is None:
            LOGGER.debug("Waiting for process with pid=%i", self._process.pid)
            self.wait()

        LOGGER.debug(
            "Process with pid=%i terminated with code=%i",
            self._process.pid,
            self._process.returncode,
        )

        self._reader.join()
        self._process.stdout.close()
        self._process.stdin.close()

    def __del__(self):
        try:
            self.terminate()
        except KeyboardInterrupt:
            LOGGER.debug("Process.__del__: Ignoring KeyboardInterrupt")


class AsynchronousFileReader(threading.Thread):
    """
    Helper class to implement asynchronous reading of a file
    in a separate thread. Pushes read lines on a queue to
    be consumed in another thread.
    """

    def __init__(self, fd, queue, encoding="utf-8"):
        threading.Thread.__init__(self)

        self._fd = io.TextIOWrapper(fd.buffer, encoding=encoding, errors="ignore")
        self._queue = queue
        self._encoding = encoding

    def run(self):
        """The body of the tread: read lines and put them on the queue."""
        for line in iter(self._fd.readline, ""):
            if PROGRAM_STATUS.is_shutting_down:
                break

            string = line[:-1]

            self._queue.put(string)
        self._queue.put(None)

    def eof(self):
        """Check whether there is no more content to expect."""
        return not self.is_alive() and self._queue.empty()


def read_file(file_name, encoding="utf-8", newline=None):
    """To stub during testing"""
    try:
        with io.open(file_name, "r", encoding=encoding, newline=newline) as file_to_read:
            data = file_to_read.read()
    except UnicodeDecodeError:
        LOGGER.warning(
            "Could not decode file %s using encoding %s, ignoring encoding errors",
            file_name,
            encoding,
        )
        with io.open(file_name, "r", encoding=encoding, errors="ignore", newline=newline) as file_to_read:
            data = file_to_read.read()

    return data


def write_file(file_name, contents, encoding="utf-8"):
    """To stub during testing"""

    path = str(Path(file_name).parent)
    if path == "":
        path = "."

    if not file_exists(path):
        makedirs(path)

    with io.open(file_name, "wb") as file_to_write:
        file_to_write.write(contents.encode(encoding=encoding))


def file_exists(file_name):
    """To stub during testing"""
    return Path(file_name).exists()


def get_modification_time(file_name):
    """To stub during testing"""
    return getmtime(file_name)


def get_time():
    """To stub during testing"""
    return time.time()


def renew_path(path):
    """
    Ensure path directory exists and is empty

    On Windows deleting a file will not actually delete it right away but only
    mark it for deletion. Therefore there is a race-condition between rmtree and makedirs.
    Virus scanners and file system indexers might temporarily block a file from being deleted right away

    http://stackoverflow.com/questions/27625683/can-anyone-explain-this-weird-behaviour-of-shutil-rmtree-and-shutil-copytree
    """
    if IS_WINDOWS_SYSTEM:
        retries = 10
        while retries > 0 and Path(path).exists():
            shutil.rmtree(path, ignore_errors=retries > 1)
            time.sleep(0.01)
            retries -= 1
    else:
        if Path(path).exists():
            shutil.rmtree(path)
    makedirs(path)


def simplify_path(path):
    """
    Return relative path towards current working directory
    unless it is a separate Windows drive
    """
    cwd = getcwd()
    drive_cwd = splitdrive(cwd)[0]
    drive_path = splitdrive(path)[0]
    if drive_path == drive_cwd:
        return relpath(path, cwd)

    return path



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
PEP8 check
"""

import unittest
from subprocess import check_call
import sys
from glob import glob
from pathlib import Path
from vunit import ROOT as RSTR

ROOT = Path(RSTR)


class TestPycodestyle(unittest.TestCase):
    """
    Test that all python code follows PEP8 Python coding standard
    """

    @staticmethod
    def test_pycodestyle():
        check_call(
            [
                sys.executable,
                "-m",
                "pycodestyle",
                "--show-source",
                "--show-pep8",
                "--max-line-length=120",
                # W503 mutually exclusive with W504
                # E722 bare except checked by pylint
                "--ignore=E402,W503,E722,E501,E203",
            ]
            + get_files_and_folders()
        )


def get_files_and_folders():
    """
    Return all files and folders which shall be arguments to pycodestyle and pylint
    """
    ret = [str(ROOT / "vunit")]
    ret += list(glob(str(ROOT / "*.py")))
    ret += list(glob(str(ROOT / "tools" / "*.py")))
    return ret



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Interface for GHDL simulator
"""

from pathlib import Path
from os import environ, makedirs, remove
import logging
import subprocess
import shlex
import re
import shutil
from json import dump
from sys import stdout  # To avoid output catched in non-verbose mode
from warnings import warn
from ..exceptions import CompileError
from ..ostools import Process
from . import SimulatorInterface, ListOfStringOption, StringOption, BooleanOption
from ..vhdl_standard import VHDL

LOGGER = logging.getLogger(__name__)


class GHDLInterface(SimulatorInterface):  # pylint: disable=too-many-instance-attributes
    """
    Interface for GHDL simulator
    """

    name = "ghdl"
    executable = environ.get("GHDL", "ghdl")
    supports_gui_flag = True
    supports_colors_in_gui = True

    compile_options = [
        ListOfStringOption("ghdl.a_flags"),
        ListOfStringOption("ghdl.flags"),
    ]

    sim_options = [
        ListOfStringOption("ghdl.sim_flags"),
        ListOfStringOption("ghdl.elab_flags"),
        StringOption("ghdl.gtkwave_script.gui"),
        BooleanOption("ghdl.elab_e"),
    ]

    @staticmethod
    def add_arguments(parser):
        """
        Add command line arguments
        """
        group = parser.add_argument_group("ghdl", description="GHDL specific flags")
        group.add_argument(
            "--gtkwave-fmt",
            choices=["vcd", "ghw"],
            default=None,
            help="Save .vcd or .ghw to open in gtkwave",
        )
        group.add_argument("--gtkwave-args", default="", help="Arguments to pass to gtkwave")

    @classmethod
    def from_args(cls, args, output_path, **kwargs):
        """
        Create instance from args namespace
        """
        prefix = cls.find_prefix()
        return cls(
            output_path=output_path,
            prefix=prefix,
            gui=args.gui,
            gtkwave_fmt=args.gtkwave_fmt,
            gtkwave_args=args.gtkwave_args,
            backend=cls.determine_backend(prefix),
        )

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find first valid ghdl toolchain prefix
        """
        return cls.find_toolchain([cls.executable])

    def __init__(  # pylint: disable=too-many-arguments
        self,
        output_path,
        prefix,
        gui=False,
        gtkwave_fmt=None,
        gtkwave_args="",
        backend="llvm",
    ):
        SimulatorInterface.__init__(self, output_path, gui)
        self._prefix = prefix
        self._project = None

        if gui and (not self.find_executable("gtkwave")):
            raise RuntimeError("Cannot find the gtkwave executable in the PATH environment variable. GUI not possible")

        self._gui = gui
        self._gtkwave_fmt = "ghw" if gui and gtkwave_fmt is None else gtkwave_fmt
        self._gtkwave_args = gtkwave_args
        self._backend = backend
        self._vhdl_standard = None
        self._coverage_test_dirs = set()

    def has_valid_exit_code(self):
        """
        Return if the simulation should fail with nonzero exit codes
        """
        return self._vhdl_standard >= VHDL.STD_2008

    @classmethod
    def _get_version_output(cls, prefix):
        """
        Get the output of 'ghdl --version'
        """
        return subprocess.check_output([str(Path(prefix) / cls.executable), "--version"]).decode()

    @classmethod
    def determine_backend(cls, prefix):
        """
        Determine the GHDL backend
        """
        mapping = {
            "mcode code generator": "mcode",
            "llvm code generator": "llvm",
            "GCC back-end code generator": "gcc",
        }
        output = cls._get_version_output(prefix)
        for name, backend in mapping.items():
            if name in output:
                LOGGER.debug("Detected GHDL %s", name)
                return backend

        LOGGER.error("Could not detect known LLVM backend by parsing 'ghdl --version'")
        print(f"Expected to find one of {mapping.keys()!r}")
        print("== Output of 'ghdl --version'" + ("=" * 60))
        print(output)
        print("=============================" + ("=" * 60))
        raise AssertionError("No known GHDL back-end could be detected from running 'ghdl --version'")

    @classmethod
    def determine_version(cls, prefix):
        """
        Determine the GHDL version
        """
        return float(
            re.match(
                r"GHDL ([0-9]*\.[0-9]*).*\(.*\) \[Dunoon edition\]",
                cls._get_version_output(prefix),
            ).group(1)
        )

    @classmethod
    def supports_vhdl_package_generics(cls):
        """
        Returns True when this simulator supports VHDL package generics
        """
        return True

    @classmethod
    def supports_vhpi(cls):
        """
        Returns True when the simulator supports VHPI
        """
        return (cls.determine_backend(cls.find_prefix_from_path()) != "mcode") or (
            cls.determine_version(cls.find_prefix_from_path()) > 0.36
        )

    @classmethod
    def supports_coverage(cls):
        """
        Returns True when the simulator supports coverage
        """
        return cls.determine_backend(cls.find_prefix_from_path()) == "gcc"

    def _has_output_flag(self):
        """
        Returns if backend supports output flag
        """
        return self._backend in ("llvm", "gcc")

    def setup_library_mapping(self, project):
        """
        Setup library mapping
        """
        self._project = project
        for library in project.get_libraries():
            if not Path(library.directory).exists():
                makedirs(library.directory)

        vhdl_standards = set(
            source_file.get_vhdl_standard()
            for source_file in project.get_source_files_in_order()
            if source_file.is_vhdl
        )

        if not vhdl_standards:
            self._vhdl_standard = VHDL.STD_2008
        elif len(vhdl_standards) != 1:
            raise RuntimeError(f"GHDL cannot handle mixed VHDL standards, found {list(vhdl_standards)!r}")
        else:
            self._vhdl_standard = list(vhdl_standards)[0]

    def compile_source_file_command(self, source_file):
        """
        Returns the command to compile a single source_file
        """
        if source_file.is_vhdl:
            return self.compile_vhdl_file_command(source_file)

        LOGGER.error("Unknown file type: %s", source_file.file_type)
        raise CompileError

    @staticmethod
    def _std_str(vhdl_standard):
        """
        Convert standard to format of GHDL command line flag
        """
        if vhdl_standard == VHDL.STD_2002:
            return "02"

        if vhdl_standard == VHDL.STD_2008:
            return "08"

        if vhdl_standard == VHDL.STD_1993:
            return "93"

        raise ValueError(f"Invalid VHDL standard {vhdl_standard!s}")

    def compile_vhdl_file_command(self, source_file):
        """
        Returns the command to compile a vhdl file
        """
        cmd = [
            str(Path(self._prefix) / self.executable),
            "-a",
            f"--workdir={source_file.library.directory!s}",
            f"--work={source_file.library.name!s}",
            f"--std={self._std_str(source_file.get_vhdl_standard())!s}",
        ]
        for library in self._project.get_libraries():
            cmd += [f"-P{library.directory!s}"]

        a_flags = source_file.compile_options.get("ghdl.a_flags", [])
        flags = source_file.compile_options.get("ghdl.flags", [])
        if flags != []:
            warn(
                ("'ghdl.flags' is deprecated and it will be removed in future releases; " "use 'ghdl.a_flags' instead"),
                Warning,
            )
            a_flags += flags

        cmd += a_flags

        if source_file.compile_options.get("enable_coverage", False):
            # Add gcc compilation flags for coverage
            #   -ftest-coverages creates .gcno notes files needed by gcov
            #   -fprofile-arcs creates branch profiling in .gcda database files
            cmd += ["-fprofile-arcs", "-ftest-coverage"]
        cmd += [source_file.name]
        return cmd

    def _get_command(self, config, output_path, elaborate_only, ghdl_e, wave_file):  # pylint: disable=too-many-branches
        """
        Return GHDL simulation command
        """
        cmd = [str(Path(self._prefix) / self.executable)]

        cmd += ["-e"] if ghdl_e else ["--elab-run"]

        cmd += [f"--std={self._std_str(self._vhdl_standard)!s}"]
        cmd += [f"--work={config.library_name!s}"]
        cmd += [f"--workdir={self._project.get_library(config.library_name).directory!s}"]
        cmd += [f"-P{lib.directory!s}" for lib in self._project.get_libraries()]

        bin_path = str(Path(output_path) / f"{config.entity_name!s}-{config.architecture_name!s}")
        if self._has_output_flag():
            cmd += ["-o", bin_path]
        cmd += config.sim_options.get("ghdl.elab_flags", [])
        if config.sim_options.get("enable_coverage", False):
            # Enable coverage in linker
            cmd += ["-Wl,-lgcov"]
        cmd += [config.entity_name, config.architecture_name]

        sim = config.sim_options.get("ghdl.sim_flags", [])
        for name, value in config.generics.items():
            sim += [f"-g{name!s}={value!s}"]
        sim += [f"--assert-level={config.vhdl_assert_stop_level!s}"]
        if config.sim_options.get("disable_ieee_warnings", False):
            sim += ["--ieee-asserts=disable"]

        if wave_file:
            if self._gtkwave_fmt == "ghw":
                sim += [f"--wave={wave_file!s}"]
            elif self._gtkwave_fmt == "vcd":
                sim += [f"--vcd={wave_file!s}"]

        if not ghdl_e:
            cmd += sim
            if elaborate_only:
                cmd += ["--no-run"]
        else:
            try:
                makedirs(output_path, mode=0o777)
            except OSError:
                pass
            with (Path(output_path) / "args.json").open("w", encoding="utf-8") as fname:
                dump(
                    {
                        "bin": str(Path(output_path) / f"{config.entity_name!s}-{config.architecture_name!s}"),
                        "build": cmd[1:],
                        "sim": sim,
                    },
                    fname,
                )

        return cmd

    def simulate(self, output_path, test_suite_name, config, elaborate_only):  # pylint: disable=too-many-locals
        """
        Simulate with entity as top level using generics
        """

        script_path = str(Path(output_path) / self.name)

        if not Path(script_path).exists():
            makedirs(script_path)

        ghdl_e = elaborate_only and config.sim_options.get("ghdl.elab_e", False)

        if self._gtkwave_fmt is not None:
            data_file_name = str(Path(script_path) / f"wave.{self._gtkwave_fmt!s}")
            if Path(data_file_name).exists():
                remove(data_file_name)
        else:
            data_file_name = None

        cmd = self._get_command(config, script_path, elaborate_only, ghdl_e, data_file_name)

        status = True

        gcov_env = environ.copy()
        if config.sim_options.get("enable_coverage", False):
            # Set environment variable to put the coverage output in the test_output folder
            coverage_dir = str(Path(output_path) / "coverage")
            gcov_env["GCOV_PREFIX"] = coverage_dir
            self._coverage_test_dirs.add(coverage_dir)

        try:
            proc = Process(cmd, env=gcov_env)
            proc.consume_output()
        except Process.NonZeroExitCode:
            status = False

        if self._gui and not elaborate_only:
            cmd = ["gtkwave"] + shlex.split(self._gtkwave_args) + [data_file_name]

            init_file = config.sim_options.get(self.name + ".gtkwave_script.gui", None)
            if init_file is not None:
                cmd += ["--script", str(Path(init_file).resolve())]

            stdout.write(" ".join(cmd) + "\n")
            subprocess.call(cmd)

        return status

    def _compile_source_file(self, source_file, printer):
        """
        Runs parent command for compilation, and moves any .gcno files to the compilation output
        """
        compilation_ok = super()._compile_source_file(source_file, printer)

        if source_file.compile_options.get("enable_coverage", False):
            # GCOV gcno files are output to where the command is run,
            # move it back to the compilation folder
            source_path = Path(source_file.name)
            gcno_file = Path(source_path.stem + ".gcno")
            if Path(gcno_file).exists():
                new_path = Path(source_file.library.directory) / gcno_file
                gcno_file.rename(new_path)

        return compilation_ok

    def merge_coverage(self, file_name, args=None):
        """
        Merge coverage from all test cases
        """
        output_dir = file_name

        # Loop over each .gcda output folder and merge them two at a time
        first_input = True
        for coverage_dir in self._coverage_test_dirs:
            if Path(coverage_dir).exists():
                merge_command = [
                    "gcov-tool",
                    "merge",
                    "-o",
                    output_dir,
                    coverage_dir if first_input else output_dir,
                    coverage_dir,
                ]
                subprocess.call(merge_command)
                first_input = False
            else:
                LOGGER.warning("Missing coverage directory: %s", coverage_dir)

        # Find actual output path of the .gcda files (they are deep in hierarchy)
        dir_path = Path(output_dir)
        gcda_dirs = {x.parent for x in dir_path.glob("**/*.gcda")}
        assert len(gcda_dirs) == 1, "Expected exactly one folder with gcda files"
        gcda_dir = gcda_dirs.pop()

        # Add compile-time .gcno files as well, they are needed for the report
        for library in self._project.get_libraries():
            for gcno_file in Path(library.directory).glob("*.gcno"):
                shutil.copy(gcno_file, gcda_dir)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Common functions re-used between test cases
"""

from pathlib import Path
from xml.etree import ElementTree
import contextlib
import functools
import os
import shutil
import random


def check_report(report_file, tests=None):
    """
    Check an XML report_file for the exact occurrence of specific test results
    """
    tree = ElementTree.parse(report_file)
    root = tree.getroot()
    report = {}
    for test in root.iter("testcase"):
        status = "passed"

        if test.find("skipped") is not None:
            status = "skipped"

        if test.find("failure") is not None:
            status = "failed"
        report[test.attrib["classname"] + "." + test.attrib["name"]] = status

    if tests is None:
        tests = []
        for name in report:
            expected_status = "failed" if "Expected to fail:" in name else "passed"
            tests.append((expected_status, name))

    for status, name in tests:
        if report[name] != status:
            raise AssertionError("Wrong status of %s got %s expected %s" % (name, report[name], status))

    num_tests = int(root.attrib["tests"])
    assert num_tests == len(tests)


def assert_exit(function, code=0):
    """
    Assert that 'function' performs SystemExit with code
    """
    try:
        function()
    except SystemExit as ex:
        assert ex.code == code
    else:
        assert False


@contextlib.contextmanager
def set_env(**environ):
    """
    Temporarily set the process environment variables.

    >>> with set_env(PLUGINS_DIR=u'test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True

    >>> "PLUGINS_DIR" in os.environ
    False

    :type environ: dict[str, unicode]
    :param environ: Environment variables to set
    """
    old_environ = dict(os.environ)
    os.environ.clear()
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


@contextlib.contextmanager
def create_tempdir(path: Path = None):
    """
    Create a temporary directory that is removed after the unit test
    """

    if path is None:
        path = Path(__file__).parent / ("tempdir_%i" % random.randint(0, 2 ** 64 - 1))

    if path.exists():
        shutil.rmtree(path)

    os.makedirs(str(path))

    try:
        yield path
    finally:
        shutil.rmtree(path)


def with_tempdir(func):
    """
    Decorator to provide test function with a temporary path that is
    removed after calling the function.

    The path is named the same as the function and its parent module
    """

    @functools.wraps(func)
    def new_function(*args, **kwargs):
        """
        Wrapper funciton that maintains temporary directory around nested
        function
        """
        path_name = Path(__file__).parent / (func.__module__ + "." + func.__name__)
        with create_tempdir(path_name) as path:
            return func(*args, tempdir=path, **kwargs)

    return new_function


def get_vhdl_test_bench(test_bench_name, tests=None, same_sim=False, test_attributes=None):
    """
    Create a valid VUnit test bench

    returns a string
    """

    if test_attributes is None:
        test_attributes = {}

    tests_contents = ""
    if tests is None:
        pass
    else:
        tests_contents = ""
        last_idx = len(tests)
        for idx, test_name in enumerate(tests):
            if idx == 0:
                tests_contents += "    if "
            else:
                tests_contents += "    elsif "

            tests_contents += 'run("%s") then\n' % test_name

            if test_name in test_attributes:
                for attr_name in test_attributes[test_name]:
                    tests_contents += "-- vunit: %s\n" % attr_name

            if idx == last_idx:
                tests_contents += "    endif;\n"

    if same_sim:
        contents = "-- vunit: run_all_in_same_sim\n"
    else:
        contents = "\n"

    contents += """\
library vunit_lib;
context vunit_lib.vunit_context;

entity {test_bench_name} is
  generic (runner_cfg : string);
end entity;

architecture a of {test_bench_name} is
begin
  main : process
  begin
    test_runner_setup(runner, runner_cfg);
    {tests_contents}
    test_runner_cleanup(runner);
  end process;
end architecture;
""".format(
        test_bench_name=test_bench_name, tests_contents=tests_contents
    )

    return contents


def create_vhdl_test_bench_file(test_bench_name, file_name, tests=None, same_sim=False, test_attributes=None):
    """
    Create a valid VUnit test bench and writes it to file_name
    """
    with Path(file_name).open("wb") as fptr:
        fptr.write(
            get_vhdl_test_bench(
                test_bench_name=test_bench_name,
                tests=tests,
                same_sim=same_sim,
                test_attributes=test_attributes,
            ).encode("utf-8")
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Simulator interface(s)
"""

import sys
import os
from os import environ, listdir, pathsep
import subprocess
from pathlib import Path
from typing import List
from ..ostools import Process, simplify_path
from ..exceptions import CompileError
from ..color_printer import NO_COLOR_PRINTER


class Option(object):
    """
    A compile or sim option
    """

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def validate(self, value):
        pass


class SimulatorInterface(object):  # pylint: disable=too-many-public-methods
    """
    Generic simulator interface
    """

    name: str = "none"
    supports_gui_flag = False
    package_users_depend_on_bodies = False
    compile_options: List[Option] = []
    sim_options: List[Option] = []

    # True if simulator supports ANSI colors in GUI mode
    supports_colors_in_gui = False

    def __init__(self, output_path, gui):
        self._output_path = output_path
        self._gui = gui

    @property
    def output_path(self):
        return self._output_path

    @property
    def use_color(self):
        return (not self._gui) or self.supports_colors_in_gui

    @staticmethod
    def add_arguments(parser):
        """
        Add command line arguments
        """

    @staticmethod
    def supports_vhdl_contexts():
        """
        Returns True when this simulator supports VHDL contexts
        """
        return True

    @staticmethod
    def find_executable(executable):
        """
        Return a list of all executables found in PATH
        """
        path = environ.get("PATH", None)
        if path is None:
            return []

        paths = path.split(pathsep)
        ext = Path(executable).suffix

        if (sys.platform == "win32" or os.name == "os2") and (ext != ".exe"):
            executable = executable + ".exe"

        result = []
        if isfile(executable):
            result.append(executable)

        for prefix in paths:
            file_name = str(Path(prefix) / executable)
            if isfile(file_name):
                # the file exists, we have a shot at spawn working
                result.append(file_name)
        return result

    @classmethod
    def find_prefix(cls):
        """
        Find prefix by looking at VUNIT_<SIMULATOR_NAME>_PATH environment variable
        """
        prefix = os.environ.get("VUNIT_" + cls.name.upper() + "_PATH", None)
        if prefix is not None:
            return prefix
        return cls.find_prefix_from_path()

    @classmethod
    def find_prefix_from_path(cls):
        """
        Find simulator toolchain prefix from PATH environment variable
        """

    @classmethod
    def is_available(cls):
        """
        Returns True if simulator is available
        """
        return cls.find_prefix() is not None

    @classmethod
    def find_toolchain(cls, executables, constraints=None):
        """
        Find the first path prefix containing all executables
        """
        constraints = [] if constraints is None else constraints

        if not executables:
            return None

        all_paths = [
            [str(Path(executables).parent.resolve()) for executables in cls.find_executable(name)]
            for name in executables
        ]

        for path0 in all_paths[0]:
            if all([path0 in paths for paths in all_paths] + [constraint(path0) for constraint in constraints]):
                return path0
        return None

    @classmethod
    def get_osvvm_coverage_api(cls):
        """
        Returns simulator name when OSVVM coverage API is supported, None otherwise.
        """

    @classmethod
    def supports_vhdl_package_generics(cls):
        """
        Returns True when this simulator supports VHDL package generics
        """
        return False

    @staticmethod
    def has_valid_exit_code():
        """
        Return if the simulation should fail with nonzero exit codes
        """
        return False

    @staticmethod
    def supports_vhpi():
        """
        Returns True when the simulator supports VHPI
        """
        return False

    @staticmethod
    def supports_coverage():
        """
        Returns True when the simulator supports coverage
        """
        return False

    def merge_coverage(self, file_name, args):  # pylint: disable=unused-argument, no-self-use
        """
        Hook for simulator interface to creating coverage reports
        """
        raise RuntimeError("This simulator does not support merging coverage")

    def add_simulator_specific(self, project):
        """
        Hook for the simulator interface to add simulator specific things to the project
        """

    def compile_project(
        self,
        project,
        printer=NO_COLOR_PRINTER,
        continue_on_error=False,
        target_files=None,
    ):
        """
        Compile the project
        param: target_files: Given a list of SourceFiles only these and dependent files are compiled
        """
        self.add_simulator_specific(project)
        self.setup_library_mapping(project)
        self.compile_source_files(project, printer, continue_on_error, target_files=target_files)

    def simulate(self, output_path, test_suite_name, config, elaborate_only):
        """
        Simulate
        """

    def setup_library_mapping(self, project):
        """
        Implemented by specific simulators
        """

    def _compile_source_file(self, source_file, printer):
        """
        Compiles a single source file and prints status information
        """
        try:
            command = self.compile_source_file_command(source_file)
        except CompileError:
            command = None
            printer.write("failed", fg="ri")
            printer.write("\n")
            printer.write(f"File type not supported by {self.name!s} simulator\n")

            return False

        try:
            output = check_output(command, env=self.get_env())
            printer.write("passed", fg="gi")
            printer.write("\n")
            printer.write(output)

        except subprocess.CalledProcessError as err:
            printer.write("failed", fg="ri")
            printer.write("\n")
            printer.write(f"=== Command used: ===\n{subprocess.list2cmdline(command)!s}\n")
            printer.write("\n")
            printer.write(f"=== Command output: ===\n{err.output!s}\n")

            return False

        return True

    def compile_source_files(
        self,
        project,
        printer=NO_COLOR_PRINTER,
        continue_on_error=False,
        target_files=None,
    ):
        """
        Use compile_source_file_command to compile all source_files
        param: target_files: Given a list of SourceFiles only these and dependent files are compiled
        """
        dependency_graph = project.create_dependency_graph()
        failures = []

        if target_files is None:
            source_files = project.get_files_in_compile_order(dependency_graph=dependency_graph)
        else:
            source_files = project.get_minimal_file_set_in_compile_order(target_files)

        source_files_to_skip = set()

        max_library_name = 0
        max_source_file_name = 0
        if source_files:
            max_library_name = max(len(source_file.library.name) for source_file in source_files)
            max_source_file_name = max(len(simplify_path(source_file.name)) for source_file in source_files)

        for source_file in source_files:
            printer.write(
                f"Compiling into {(source_file.library.name + ':').ljust(max_library_name + 1)!s} "
                f"{simplify_path(source_file.name).ljust(max_source_file_name)!s} "
            )
            sys.stdout.flush()

            if source_file in source_files_to_skip:
                printer.write("skipped", fg="rgi")
                printer.write("\n")
                continue

            if self._compile_source_file(source_file, printer):
                project.update(source_file)
            else:
                source_files_to_skip.update(dependency_graph.get_dependent([source_file]))
                failures.append(source_file)

                if not continue_on_error:
                    break

        if failures:
            printer.write("Compile failed\n", fg="ri")
            raise CompileError

        if source_files:
            printer.write("Compile passed\n", fg="gi")
        else:
            printer.write("Re-compile not needed\n")

    def compile_source_file_command(self, source_file):  # pylint: disable=unused-argument
        raise NotImplementedError

    @staticmethod
    def get_env():
        """
        Allows inheriting classes to overload this to modify environment variables. Return None for default environment
        """


def isfile(file_name):
    """
    Case insensitive Path.is_file()
    """
    fpath = Path(file_name)
    if not fpath.is_file():
        return False

    return str(fpath.name) in listdir(str(fpath.parent))


def run_command(command, cwd=None, env=None):
    """
    Run a command
    """
    try:
        proc = Process(command, cwd=cwd, env=env)
        proc.consume_output()
        return True
    except Process.NonZeroExitCode:
        pass
    return False


def check_output(command, env=None):
    """
    Wrapper arround subprocess.check_output
    """
    try:
        output = subprocess.check_output(  # pylint: disable=unexpected-keyword-arg
            command, env=env, stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as err:
        err.output = err.output.decode("utf-8")
        raise err
    return output.decode("utf-8")


class BooleanOption(Option):
    """
    Must be a boolean
    """

    def validate(self, value):
        if value not in (True, False):
            raise ValueError(f"Option {self.name!r} must be a boolean. Got {value!r}")


class StringOption(Option):
    """
    Must be a string
    """

    def validate(self, value):
        if not is_string_not_iterable(value):
            raise ValueError(f"Option {self.name!r} must be a string. Got {value!r}")


class ListOfStringOption(Option):
    """
    Must be a list of strings
    """

    def validate(self, value):
        def fail():
            raise ValueError(f"Option {self.name!r} must be a list of strings. Got {value!r}")

        if is_string_not_iterable(value):
            fail()

        try:
            for elem in value:
                if not is_string_not_iterable(elem):
                    fail()
        except TypeError:
            fail()


class VHDLAssertLevelOption(Option):
    """
    VHDL assert level
    """

    _legal_values = ("warning", "error", "failure")

    def __init__(self):
        Option.__init__(self, "vhdl_assert_stop_level")

    def validate(self, value):
        if value not in self._legal_values:
            raise ValueError(f"Option {self.name!r} must be one of {self._legal_values!s}. Got {value!r}")


def is_string_not_iterable(value):
    """
    Returns True if value is a string and not another iterable
    """
    return isinstance(value, str)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

from os import getenv
import glob
from pathlib import Path
from vunit import VUnit, location_preprocessor


def main():
    vhdl_2019 = getenv("VUNIT_VHDL_STANDARD") == "2019"
    root = Path(__file__).parent

    ui = VUnit.from_argv()

    vunit_lib = ui.library("vunit_lib")
    files = glob.glob(str(root / "test" / "*.vhd"))
    files.remove(str(root / "test" / "tb_location.vhd"))
    vunit_lib.add_source_files(files)

    preprocessor = location_preprocessor.LocationPreprocessor()
    preprocessor.add_subprogram("print_pre_vhdl_2019_style")
    preprocessor.remove_subprogram("info")
    vunit_lib.add_source_files(root / "test" / "tb_location.vhd", preprocessors=[preprocessor])

    if vhdl_2019:
        testbenches = vunit_lib.get_source_files("*tb*")
        testbenches.set_compile_option("rivierapro.vcom_flags", ["-dbg"])
        ui.set_sim_option("rivierapro.vsim_flags", ["-filter RUNTIME_0375"])

    vunit_lib.test_bench("tb_location").set_generic("vhdl_2019", vhdl_2019)

    ui.main()


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Interface towards Aldec Active HDL
"""

from functools import total_ordering
from pathlib import Path
import os
import re
import logging
from ..exceptions import CompileError
from ..ostools import Process, write_file, file_exists, renew_path
from ..vhdl_standard import VHDL
from ..test.suites import get_result_file_name
from . import SimulatorInterface, ListOfStringOption, StringOption
from .vsim_simulator_mixin import get_is_test_suite_done_tcl, fix_path

LOGGER = logging.getLogger(__name__)


class ActiveHDLInterface(SimulatorInterface):
    """
    Active HDL interface
    """

    name = "activehdl"
    supports_gui_flag = True
    package_users_depend_on_bodies = True
    compile_options = [
        ListOfStringOption("activehdl.vcom_flags"),
        ListOfStringOption("activehdl.vlog_flags"),
    ]

    sim_options = [
        ListOfStringOption("activehdl.vsim_flags"),
        ListOfStringOption("activehdl.vsim_flags.gui"),
        StringOption("activehdl.init_file.gui"),
    ]

    @classmethod
    def from_args(cls, args, output_path, **kwargs):
        """
        Create new instance from command line arguments object
        """
        return cls(prefix=cls.find_prefix(), output_path=output_path, gui=args.gui)

    @classmethod
    def find_prefix_from_path(cls):
        return cls.find_toolchain(["vsim", "avhdl"])

    @classmethod
    def supports_vhdl_package_generics(cls):
        """
        Returns True when this simulator supports VHDL package generics
        """
        proc = Process([str(Path(cls.find_prefix()) / "vcom"), "-version"], env=cls.get_env())
        consumer = VersionConsumer()
        proc.consume_output(consumer)
        if consumer.version is not None:
            return consumer.version >= Version(10, 1)

        return False

    @staticmethod
    def supports_coverage():
        """
        Returns True when the simulator supports coverage
        """
        return True

    def __init__(self, prefix, output_path, gui=False):
        SimulatorInterface.__init__(self, output_path, gui)
        self._library_cfg = str(Path(output_path) / "library.cfg")
        self._prefix = prefix
        self._create_library_cfg()
        self._libraries = []
        self._coverage_files = set()

    def setup_library_mapping(self, project):
        """
        Setup library mapping
        """
        mapped_libraries = self._get_mapped_libraries()

        for library in project.get_libraries():
            self._libraries.append(library)
            self.create_library(library.name, library.directory, mapped_libraries)

    def compile_source_file_command(self, source_file):
        """
        Returns the command to compile a single source_file
        """
        if source_file.is_vhdl:
            return self.compile_vhdl_file_command(source_file)

        if source_file.is_any_verilog:
            return self.compile_verilog_file_command(source_file)

        LOGGER.error("Unknown file type: %s", source_file.file_type)
        raise CompileError

    @staticmethod
    def _std_str(vhdl_standard):
        """
        Convert standard to format of Active-HDL command line flag
        """
        if vhdl_standard <= VHDL.STD_2008:
            return f"-{vhdl_standard!s}"

        raise ValueError(f"Invalid VHDL standard {vhdl_standard!s}")

    def compile_vhdl_file_command(self, source_file):
        """
        Returns the command to compile a VHDL file
        """
        return (
            [
                str(Path(self._prefix) / "vcom"),
                "-quiet",
                "-j",
                str(Path(self._library_cfg).parent),
            ]
            + source_file.compile_options.get("activehdl.vcom_flags", [])
            + [
                self._std_str(source_file.get_vhdl_standard()),
                "-work",
                source_file.library.name,
                source_file.name,
            ]
        )

    def compile_verilog_file_command(self, source_file):
        """
        Returns the command to compile a Verilog file
        """
        args = [str(Path(self._prefix) / "vlog"), "-quiet", "-lc", self._library_cfg]
        args += source_file.compile_options.get("activehdl.vlog_flags", [])
        args += ["-work", source_file.library.name, source_file.name]
        for library in self._libraries:
            args += ["-l", library.name]
        for include_dir in source_file.include_dirs:
            args += [f"+incdir+{include_dir!s}"]
        for key, value in source_file.defines.items():
            args += [f"+define+{key!s}={value!s}"]
        return args

    def create_library(self, library_name, path, mapped_libraries=None):
        """
        Create and map a library_name to path
        """
        mapped_libraries = mapped_libraries if mapped_libraries is not None else {}

        apath = str(Path(path).parent.resolve())

        if not file_exists(apath):
            os.makedirs(apath)

        if not file_exists(path):
            proc = Process(
                [str(Path(self._prefix) / "vlib"), library_name, path],
                cwd=str(Path(self._library_cfg).parent),
                env=self.get_env(),
            )
            proc.consume_output(callback=None)

        if library_name in mapped_libraries and mapped_libraries[library_name] == path:
            return

        proc = Process(
            [str(Path(self._prefix) / "vmap"), library_name, path],
            cwd=str(Path(self._library_cfg).parent),
            env=self.get_env(),
        )
        proc.consume_output(callback=None)

    def _create_library_cfg(self):
        """
        Create the library.cfg file if it does not exist
        """
        if file_exists(self._library_cfg):
            return

        with Path(self._library_cfg).open("w", encoding="utf-8") as ofile:
            ofile.write(f'$INCLUDE = "{str(Path(self._prefix).parent / "vlib" / "library.cfg")}"\n')

    _library_re = re.compile(r'([a-zA-Z_]+)\s=\s"(.*)"')

    def _get_mapped_libraries(self):
        """
        Get mapped libraries from library.cfg file
        """
        with Path(self._library_cfg).open("r", encoding="utf-8") as fptr:
            text = fptr.read()

        libraries = {}
        for line in text.splitlines():
            match = self._library_re.match(line)
            if match is None:
                continue
            key = match.group(1)
            value = match.group(2)
            libraries[key] = str((Path(self._library_cfg).parent / Path(value).parent).resolve())
        return libraries

    def _vsim_extra_args(self, config):
        """
        Determine vsim_extra_args
        """
        vsim_extra_args = []
        vsim_extra_args = config.sim_options.get("activehdl.vsim_flags", vsim_extra_args)

        if self._gui:
            vsim_extra_args = config.sim_options.get("activehdl.vsim_flags.gui", vsim_extra_args)

        return " ".join(vsim_extra_args)

    def _create_load_function(self, config, output_path):
        """
        Create the vunit_load TCL function that runs the vsim command and loads the design
        """
        set_generic_str = "\n    ".join(
            (f"set vunit_generic_{name!s} {{{value!s}}}" for name, value in config.generics.items())
        )
        set_generic_name_str = " ".join(
            (f"-g/{config.entity_name!s}/{name!s}=${{vunit_generic_{name!s}}}" for name in config.generics)
        )
        pli_str = " ".join(f'-pli "{fix_path(name)}"' for name in config.sim_options.get("pli", []))

        vsim_flags = [
            pli_str,
            set_generic_name_str,
            "-lib",
            config.library_name,
            config.entity_name,
        ]

        if config.architecture_name is not None:
            vsim_flags.append(config.architecture_name)

        if config.sim_options.get("enable_coverage", False):
            coverage_file_path = str(Path(output_path) / "coverage.acdb")
            self._coverage_files.add(coverage_file_path)
            vsim_flags += [f"-acdb_file {{{fix_path(coverage_file_path)!s}}}"]

        vsim_flags += [self._vsim_extra_args(config)]

        if config.sim_options.get("disable_ieee_warnings", False):
            vsim_flags.append("-ieee_nowarn")

        # Add the the testbench top-level unit last as coverage is
        # only collected for the top-level unit specified last

        vhdl_assert_stop_level_mapping = dict(warning=1, error=2, failure=3)

        tcl = f"""
proc vunit_load {{}} {{
    {set_generic_str}
    set vsim_failed [catch {{
        vsim {' '.join(vsim_flags)}
    }}]
    if {{${{vsim_failed}}}} {{
        return true
    }}

    global breakassertlevel
    set breakassertlevel {{{vhdl_assert_stop_level_mapping[config.vhdl_assert_stop_level]}}}

    global builtinbreakassertlevel
    set builtinbreakassertlevel $breakassertlevel

    return false
}}
"""

        return tcl

    @staticmethod
    def _create_run_function():
        """
        Create the vunit_run function to run the test bench
        """
        return """
proc vunit_run {} {
    run -all
    if {![is_test_suite_done]} {
        catch {
            # tb command can fail when error comes from pli
            echo ""
            echo "Stack trace result from 'bt' command"
            bt
        }
        return true;
    }
    return false;
}
"""

    def merge_coverage(self, file_name, args=None):
        """
        Merge coverage from all test cases,
        """

        merge_command = "onerror {quit -code 1}\n"
        merge_command += "acdb merge"

        for coverage_file in self._coverage_files:
            if file_exists(coverage_file):
                merge_command += f" -i {{{fix_path(coverage_file)!s}}}"
            else:
                LOGGER.warning("Missing coverage file: %s", coverage_file)

        if args is not None:
            merge_command += " " + " ".join(f"{{{arg!s}}}" for arg in args)

        merge_command += f" -o {{{fix_path(file_name)!s}}}\n"

        merge_script_name = str(Path(self._output_path) / "acdb_merge.tcl")
        with Path(merge_script_name).open("w", encoding="utf-8") as fptr:
            fptr.write(merge_command + "\n")

        vcover_cmd = [
            str(Path(self._prefix) / "vsimsa"),
            "-tcl",
            str(fix_path(merge_script_name)),
        ]

        print(f"Merging coverage files into {file_name!s}...")
        vcover_merge_process = Process(vcover_cmd, env=self.get_env())
        vcover_merge_process.consume_output()
        print("Done merging coverage files")

    def _create_common_script(self, config, output_path):
        """
        Create tcl script with functions common to interactive and batch modes
        """
        tcl = ""
        tcl += get_is_test_suite_done_tcl(get_result_file_name(output_path))
        tcl += self._create_load_function(config, output_path)
        tcl += self._create_run_function()
        return tcl

    @staticmethod
    def _create_batch_script(common_file_name, load_only=False):
        """
        Create tcl script to run in batch mode
        """
        batch_do = ""
        batch_do += f'source "{fix_path(common_file_name)!s}"\n'
        batch_do += "set failed [vunit_load]\n"
        batch_do += "if {$failed} {quit -code 1}\n"
        if not load_only:
            batch_do += "set failed [vunit_run]\n"
            batch_do += "if {$failed} {quit -code 1}\n"
        batch_do += "quit -code 0\n"
        return batch_do

    def _create_gui_script(self, common_file_name, config):
        """
        Create the user facing script which loads common functions and prints a help message
        """

        tcl = ""
        tcl += f'source "{fix_path(common_file_name)!s}"\n'
        tcl += "workspace create workspace\n"
        tcl += "design create -a design .\n"

        for library in self._libraries:
            tcl += f"vmap {library.name!s} {fix_path(library.directory)!s}\n"

        tcl += "vunit_load\n"

        init_file = config.sim_options.get(self.name + ".init_file.gui", None)
        if init_file is not None:
            tcl += f'source "{fix_path(str(Path(init_file).resolve()))!s}"\n'

        tcl += 'puts "VUnit help: Design already loaded. Use run -all to run the test."\n'

        return tcl

    def _run_batch_file(self, batch_file_name, gui, cwd):
        """
        Run a test bench in batch by invoking a new vsim process from the command line
        """

        todo = f'@do -tcl ""{fix_path(batch_file_name)!s}""'
        if not gui:
            todo = "@onerror {quit -code 1};" + todo

        try:
            args = [
                str(Path(self._prefix) / "vsim"),
                "-gui" if gui else "-c",
                "-l",
                str(Path(batch_file_name).parent / "transcript"),
                "-do",
                todo,
            ]

            proc = Process(args, cwd=cwd, env=self.get_env())
            proc.consume_output()
        except Process.NonZeroExitCode:
            return False
        return True

    def simulate(self, output_path, test_suite_name, config, elaborate_only):
        """
        Run a test bench
        """
        script_path = Path(output_path) / self.name
        common_file_name = script_path / "common.tcl"
        batch_file_name = script_path / "batch.tcl"
        gui_file_name = script_path / "gui.tcl"

        write_file(common_file_name, self._create_common_script(config, output_path))
        write_file(gui_file_name, self._create_gui_script(str(common_file_name), config))
        write_file(
            str(batch_file_name),
            self._create_batch_script(str(common_file_name), elaborate_only),
        )

        if self._gui:
            gui_path = str(script_path / "gui")
            renew_path(gui_path)
            return self._run_batch_file(str(gui_file_name), gui=True, cwd=gui_path)

        return self._run_batch_file(str(batch_file_name), gui=False, cwd=str(Path(self._library_cfg).parent))


@total_ordering
class Version(object):
    """
    Simulator version
    """

    def __init__(self, major=0, minor=0, minor_letter=""):
        self.major = major
        self.minor = minor
        self.minor_letter = minor_letter

    def _compare(self, other, greater_than, less_than, equal_to):
        """
        Compares this object with another
        """
        if self.major > other.major:
            result = greater_than
        elif self.major < other.major:
            result = less_than
        elif self.minor > other.minor:
            result = greater_than
        elif self.minor < other.minor:
            result = less_than
        elif self.minor_letter > other.minor_letter:
            result = greater_than
        elif self.minor_letter < other.minor_letter:
            result = less_than
        else:
            result = equal_to

        return result

    def __lt__(self, other):
        return self._compare(other, greater_than=False, less_than=True, equal_to=False)

    def __eq__(self, other):
        return self._compare(other, greater_than=False, less_than=False, equal_to=True)


class VersionConsumer(object):
    """
    Consume version information
    """

    def __init__(self):
        self.version = None

    _version_re = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)(?P<minor_letter>[a-zA-Z]?)\.\d+\.\d+")

    def __call__(self, line):
        match = self._version_re.search(line)
        if match is not None:
            major = int(match.group("major"))
            minor = int(match.group("minor"))
            minor_letter = match.group("minor_letter")
            self.version = Version(major, minor, minor_letter)
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2021, Lars Asplund lars.anders.asplund@gmail.com

"""
Test the vhdl_standard.py file
"""

from vunit.vhdl_standard import VHDL


def test_valid_standards():
    for std in ["93", "02", "08", "19", "1993", "2002", "2008", "2019"]:
        VHDL.standard(std)


def test_error_on_invalid_standard():
    _assert_is_invalid("2001")
    _assert_is_invalid("002")
    _assert_is_invalid("993")
    _assert_is_invalid("2")
    _assert_is_invalid("3")


def test_equality():
    assert VHDL.standard("2008") == VHDL.standard("2008")
    assert VHDL.standard("1993") != VHDL.standard("2008")
    assert VHDL.standard("93") == VHDL.standard("1993")


def test_comparison():
    assert VHDL.standard("1993") < VHDL.standard("2002")
    assert VHDL.standard("2002") < VHDL.standard("2008")
    assert VHDL.standard("2008") < VHDL.standard("2019")


def test_str():
    assert str(VHDL.standard("1993")) == "93"
    assert str(VHDL.standard("2002")) == "2002"


def test_and_later():
    assert VHDL.STD_1993.and_later == {
        VHDL.STD_1993,
        VHDL.STD_2002,
        VHDL.STD_2008,
        VHDL.STD_2019,
    }
    assert VHDL.STD_2002.and_later == {
        VHDL.STD_2002,
        VHDL.STD_2008,
        VHDL.STD_2019,
    }
    assert VHDL.STD_2008.and_later == {VHDL.STD_2008, VHDL.STD_2019}
    assert VHDL.STD_2019.and_later == {VHDL.STD_2019}


def test_and_earlier():
    assert VHDL.STD_2019.and_earlier == {
        VHDL.STD_1993,
        VHDL.STD_2002,
        VHDL.STD_2008,
        VHDL.STD_2019,
    }
    assert VHDL.STD_2008.and_earlier == {VHDL.STD_1993, VHDL.STD_2002, VHDL.STD_2008}
    assert VHDL.STD_2002.and_earlier == {VHDL.STD_1993, VHDL.STD_2002}
    assert VHDL.STD_1993.and_earlier == {VHDL.STD_1993}


def test_supports_context():
    assert not VHDL.STD_2002.supports_context
    assert VHDL.STD_2008.supports_context


def _assert_is_invalid(standard_string):
    """
    Check that the standard string produces an exception
    """
    try:
        VHDL.standard(standard_string)
    except ValueError:
        pass
    else:
        raise AssertionError("Exception not raised")
